#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG that uses Google PubSub services.
"""

from __future__ import annotations

import os
from datetime import datetime

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.pubsub import (
    PubSubCreateSubscriptionOperator,
    PubSubCreateTopicOperator,
    PubSubDeleteSubscriptionOperator,
    PubSubDeleteTopicOperator,
    PubSubPublishMessageOperator,
    PubSubPullOperator,
)
from airflow.providers.google.cloud.sensors.pubsub import PubSubPullSensor
from airflow.providers.standard.operators.bash import BashOperator

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT", "default")

DAG_ID = "pubsub"

TOPIC_ID = f"topic-{DAG_ID}-{ENV_ID}"
MESSAGE = {"data": b"Tool", "attributes": {"name": "wrench", "mass": "1.3kg", "count": "3"}}
MESSAGE_TWO = {"data": b"Tool", "attributes": {"name": "wrench", "mass": "1.2kg", "count": "2"}}

# [START howto_operator_gcp_pubsub_pull_messages_result_cmd]
echo_cmd = """
{% for m in task_instance.xcom_pull('pull_messages') %}
    echo "AckID: {{ m.get('ackId') }}, Base64-Encoded: {{ m.get('message') }}"
{% endfor %}
"""
# [END howto_operator_gcp_pubsub_pull_messages_result_cmd]

with DAG(
    DAG_ID,
    schedule="@once",  # Override to match your needs
    start_date=datetime(2021, 1, 1),
    catchup=False,
) as dag:
    # [START howto_operator_gcp_pubsub_create_topic]
    create_topic = PubSubCreateTopicOperator(
        task_id="create_topic", topic=TOPIC_ID, project_id=PROJECT_ID, fail_if_exists=False
    )
    # [END howto_operator_gcp_pubsub_create_topic]

    # [START howto_operator_gcp_pubsub_create_subscription]
    subscribe_task = PubSubCreateSubscriptionOperator(
        task_id="subscribe_task", project_id=PROJECT_ID, topic=TOPIC_ID
    )
    # [END howto_operator_gcp_pubsub_create_subscription]

    # [START howto_operator_gcp_pubsub_pull_message_with_sensor]
    subscription = subscribe_task.output

    pull_messages = PubSubPullSensor(
        task_id="pull_messages",
        ack_messages=True,
        project_id=PROJECT_ID,
        subscription=subscription,
    )
    # [END howto_operator_gcp_pubsub_pull_message_with_sensor]

    # [START howto_operator_gcp_pubsub_pull_messages_result]
    pull_messages_result = BashOperator(task_id="pull_messages_result", bash_command=echo_cmd)
    # [END howto_operator_gcp_pubsub_pull_messages_result]

    # [START howto_operator_gcp_pubsub_pull_message_with_operator]

    pull_messages_operator = PubSubPullOperator(
        task_id="pull_messages_operator",
        ack_messages=True,
        project_id=PROJECT_ID,
        subscription=subscription,
    )
    # [END howto_operator_gcp_pubsub_pull_message_with_operator]

    # [START howto_operator_gcp_pubsub_publish]
    publish_task = PubSubPublishMessageOperator(
        task_id="publish_task",
        project_id=PROJECT_ID,
        topic=TOPIC_ID,
        messages=[MESSAGE, MESSAGE],
    )
    # [END howto_operator_gcp_pubsub_publish]

    publish_task2 = PubSubPublishMessageOperator(
        task_id="publish_task2",
        project_id=PROJECT_ID,
        topic=TOPIC_ID,
        messages=[MESSAGE_TWO, MESSAGE_TWO],
    )

    # [START howto_operator_gcp_pubsub_unsubscribe]
    unsubscribe_task = PubSubDeleteSubscriptionOperator(
        task_id="unsubscribe_task",
        project_id=PROJECT_ID,
        subscription=subscription,
    )
    # [END howto_operator_gcp_pubsub_unsubscribe]
    unsubscribe_task.trigger_rule = TriggerRule.ALL_DONE

    # [START howto_operator_gcp_pubsub_delete_topic]
    delete_topic = PubSubDeleteTopicOperator(task_id="delete_topic", topic=TOPIC_ID, project_id=PROJECT_ID)
    # [END howto_operator_gcp_pubsub_delete_topic]
    delete_topic.trigger_rule = TriggerRule.ALL_DONE

    (
        create_topic
        >> subscribe_task
        >> publish_task
        >> pull_messages
        >> pull_messages_result
        >> publish_task2
        >> pull_messages_operator
        >> unsubscribe_task
        >> delete_topic
    )

    # Task dependencies created via `XComArgs`:
    #   subscribe_task >> pull_messages
    #   subscribe_task >> pull_messages_operator
    #   subscribe_task >> unsubscribe_task

    # ### Everything below this line is not part of example ###
    # ### Just for system tests purpose ###
    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


"""
Example Airflow DAG for Google Vertex AI service testing Model Service operations.
"""

from __future__ import annotations

import os
from datetime import datetime

from google.cloud.aiplatform import schema
from google.protobuf.json_format import ParseDict
from google.protobuf.struct_pb2 import Value

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.gcs import (
    GCSCreateBucketOperator,
    GCSDeleteBucketOperator,
    GCSSynchronizeBucketsOperator,
)
from airflow.providers.google.cloud.operators.vertex_ai.custom_job import (
    CreateCustomTrainingJobOperator,
    DeleteCustomTrainingJobOperator,
)
from airflow.providers.google.cloud.operators.vertex_ai.dataset import (
    CreateDatasetOperator,
    DeleteDatasetOperator,
)
from airflow.providers.google.cloud.operators.vertex_ai.model_service import (
    AddVersionAliasesOnModelOperator,
    DeleteModelOperator,
    DeleteModelVersionOperator,
    DeleteVersionAliasesOnModelOperator,
    ExportModelOperator,
    GetModelOperator,
    ListModelsOperator,
    ListModelVersionsOperator,
    SetDefaultVersionOnModelOperator,
    UploadModelOperator,
)
from airflow.providers.google.cloud.transfers.gcs_to_local import GCSToLocalFilesystemOperator

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT", "default")
DAG_ID = "vertex_ai_model_service_operations"
REGION = "us-central1"
TRAIN_DISPLAY_NAME = f"train-housing-custom-{ENV_ID}"
MODEL_DISPLAY_NAME = f"custom-housing-model-{ENV_ID}"

RESOURCE_DATA_BUCKET = "airflow-system-tests-resources"
DATA_SAMPLE_GCS_BUCKET_NAME = f"bucket_{DAG_ID}_{ENV_ID}".replace("_", "-")
STAGING_BUCKET = f"gs://{DATA_SAMPLE_GCS_BUCKET_NAME}"

DATA_SAMPLE_GCS_OBJECT_NAME = "vertex-ai/california_housing_train.csv"

TABULAR_DATASET = {
    "display_name": f"tabular-dataset-{ENV_ID}",
    "metadata_schema_uri": schema.dataset.metadata.tabular,
    "metadata": ParseDict(
        {
            "input_config": {
                "gcs_source": {"uri": [f"gs://{DATA_SAMPLE_GCS_BUCKET_NAME}/{DATA_SAMPLE_GCS_OBJECT_NAME}"]}
            }
        },
        Value(),
    ),
}

CONTAINER_URI = "us-docker.pkg.dev/vertex-ai/training/tf-cpu.2-2:latest"

# LOCAL_TRAINING_SCRIPT_PATH should be set for Airflow which is running on distributed system.
# For example in Composer the correct path is `gcs/data/california_housing_training_script.py`.
# Because `gcs/data/` is shared folder for Airflow's workers.
IS_COMPOSER = bool(os.environ.get("COMPOSER_ENVIRONMENT", ""))
LOCAL_TRAINING_SCRIPT_PATH = (
    "gcs/data/california_housing_training_script.py"
    if IS_COMPOSER
    else "california_housing_training_script.py"
)

MODEL_OUTPUT_CONFIG = {
    "artifact_destination": {
        "output_uri_prefix": STAGING_BUCKET,
    },
    "export_format_id": "custom-trained",
}
MODEL_SERVING_CONTAINER_URI = "us-docker.pkg.dev/vertex-ai/prediction/tf2-cpu.2-2:latest"
MODEL_OBJ = {
    "display_name": f"model-{ENV_ID}",
    "artifact_uri": "{{ti.xcom_pull('custom_task')['artifactUri']}}",
    "container_spec": {
        "image_uri": MODEL_SERVING_CONTAINER_URI,
        "command": [],
        "args": [],
        "env": [],
        "ports": [],
        "predict_route": "",
        "health_route": "",
    },
}
MODEL_OBJ_V2 = {
    "display_name": f"model-{ENV_ID}-v2",
    "artifact_uri": "{{ti.xcom_pull('custom_task')['artifactUri']}}",
    "container_spec": {
        "image_uri": MODEL_SERVING_CONTAINER_URI,
        "command": [],
        "args": [],
        "env": [],
        "ports": [],
        "predict_route": "",
        "health_route": "",
    },
}


with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    render_template_as_native_obj=True,
    tags=["example", "vertex_ai", "model_service"],
) as dag:
    create_bucket = GCSCreateBucketOperator(
        task_id="create_bucket",
        bucket_name=DATA_SAMPLE_GCS_BUCKET_NAME,
        storage_class="REGIONAL",
        location=REGION,
    )

    move_data_files = GCSSynchronizeBucketsOperator(
        task_id="move_files_to_bucket",
        source_bucket=RESOURCE_DATA_BUCKET,
        source_object="vertex-ai/california-housing-data",
        destination_bucket=DATA_SAMPLE_GCS_BUCKET_NAME,
        destination_object="vertex-ai",
        recursive=True,
    )

    download_training_script_file = GCSToLocalFilesystemOperator(
        task_id="download_training_script_file",
        object_name="vertex-ai/california_housing_training_script.py",
        bucket=DATA_SAMPLE_GCS_BUCKET_NAME,
        filename=LOCAL_TRAINING_SCRIPT_PATH,
    )

    create_tabular_dataset = CreateDatasetOperator(
        task_id="tabular_dataset",
        dataset=TABULAR_DATASET,
        region=REGION,
        project_id=PROJECT_ID,
    )
    tabular_dataset_id = create_tabular_dataset.output["dataset_id"]

    create_custom_training_job = CreateCustomTrainingJobOperator(
        task_id="custom_task",
        staging_bucket=f"gs://{DATA_SAMPLE_GCS_BUCKET_NAME}",
        display_name=TRAIN_DISPLAY_NAME,
        script_path=LOCAL_TRAINING_SCRIPT_PATH,
        container_uri=CONTAINER_URI,
        requirements=["gcsfs==0.7.1"],
        model_serving_container_image_uri=MODEL_SERVING_CONTAINER_URI,
        # run params
        dataset_id=tabular_dataset_id,
        replica_count=1,
        model_display_name=MODEL_DISPLAY_NAME,
        region=REGION,
        project_id=PROJECT_ID,
    )
    model_id_v1 = create_custom_training_job.output["model_id"]

    create_custom_training_job_v2 = CreateCustomTrainingJobOperator(
        task_id="custom_task_v2",
        staging_bucket=f"gs://{DATA_SAMPLE_GCS_BUCKET_NAME}",
        display_name=TRAIN_DISPLAY_NAME,
        script_path=LOCAL_TRAINING_SCRIPT_PATH,
        container_uri=CONTAINER_URI,
        requirements=["gcsfs==0.7.1"],
        model_serving_container_image_uri=MODEL_SERVING_CONTAINER_URI,
        parent_model=model_id_v1,
        # run params
        dataset_id=tabular_dataset_id,
        replica_count=1,
        model_display_name=MODEL_DISPLAY_NAME,
        region=REGION,
        project_id=PROJECT_ID,
    )
    model_id_v2 = create_custom_training_job_v2.output["model_id"]

    # [START how_to_cloud_vertex_ai_get_model_operator]
    get_model = GetModelOperator(
        task_id="get_model", region=REGION, project_id=PROJECT_ID, model_id=model_id_v1
    )
    # [END how_to_cloud_vertex_ai_get_model_operator]

    # [START how_to_cloud_vertex_ai_list_model_versions_operator]
    list_model_versions = ListModelVersionsOperator(
        task_id="list_model_versions", region=REGION, project_id=PROJECT_ID, model_id=model_id_v1
    )
    # [END how_to_cloud_vertex_ai_list_model_versions_operator]

    # [START how_to_cloud_vertex_ai_set_version_as_default_operator]
    set_default_version = SetDefaultVersionOnModelOperator(
        task_id="set_default_version",
        project_id=PROJECT_ID,
        region=REGION,
        model_id=model_id_v2,
    )
    # [END how_to_cloud_vertex_ai_set_version_as_default_operator]

    # [START how_to_cloud_vertex_ai_add_version_aliases_operator]
    add_version_alias = AddVersionAliasesOnModelOperator(
        task_id="add_version_alias",
        project_id=PROJECT_ID,
        region=REGION,
        version_aliases=["new-version", "beta"],
        model_id=model_id_v2,
    )
    # [END how_to_cloud_vertex_ai_add_version_aliases_operator]

    # [START how_to_cloud_vertex_ai_upload_model_operator]
    upload_model = UploadModelOperator(
        task_id="upload_model",
        region=REGION,
        project_id=PROJECT_ID,
        model=MODEL_OBJ,
    )
    upload_model_v1 = upload_model.output["model_id"]
    # [END how_to_cloud_vertex_ai_upload_model_operator]
    upload_model_with_parent_model = UploadModelOperator(
        task_id="upload_model_with_parent_model",
        region=REGION,
        project_id=PROJECT_ID,
        model=MODEL_OBJ_V2,
        parent_model=upload_model_v1,
    )

    # [START how_to_cloud_vertex_ai_export_model_operator]
    export_model = ExportModelOperator(
        task_id="export_model",
        project_id=PROJECT_ID,
        region=REGION,
        model_id=upload_model.output["model_id"],
        output_config=MODEL_OUTPUT_CONFIG,
    )
    # [END how_to_cloud_vertex_ai_export_model_operator]

    # [START how_to_cloud_vertex_ai_delete_model_operator]
    delete_model = DeleteModelOperator(
        task_id="delete_model",
        project_id=PROJECT_ID,
        region=REGION,
        model_id=upload_model.output["model_id"],
        trigger_rule=TriggerRule.ALL_DONE,
    )
    # [END how_to_cloud_vertex_ai_delete_model_operator]
    delete_model_with_parent_model = DeleteModelOperator(
        task_id="delete_model_with_parent_model",
        project_id=PROJECT_ID,
        region=REGION,
        model_id=upload_model_with_parent_model.output["model_id"],
        trigger_rule=TriggerRule.ALL_DONE,
    )

    # [START how_to_cloud_vertex_ai_list_models_operator]
    list_models = ListModelsOperator(
        task_id="list_models",
        region=REGION,
        project_id=PROJECT_ID,
    )
    # [END how_to_cloud_vertex_ai_list_models_operator]

    # [START how_to_cloud_vertex_ai_delete_version_aliases_operator]
    delete_version_alias = DeleteVersionAliasesOnModelOperator(
        task_id="delete_version_alias",
        project_id=PROJECT_ID,
        region=REGION,
        version_aliases=["new-version"],
        model_id=model_id_v2,
    )
    # [END how_to_cloud_vertex_ai_delete_version_aliases_operator]

    # [START how_to_cloud_vertex_ai_delete_version_operator]
    delete_model_version = DeleteModelVersionOperator(
        task_id="delete_model_version",
        project_id=PROJECT_ID,
        region=REGION,
        model_id=model_id_v1,
        trigger_rule=TriggerRule.ALL_DONE,
    )
    # [END how_to_cloud_vertex_ai_delete_version_operator]

    delete_custom_training_job = DeleteCustomTrainingJobOperator(
        task_id="delete_custom_training_job",
        training_pipeline_id="{{ task_instance.xcom_pull(task_ids='custom_task', key='training_id') }}",
        custom_job_id="{{ task_instance.xcom_pull(task_ids='custom_task', key='custom_job_id') }}",
        region=REGION,
        project_id=PROJECT_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_tabular_dataset = DeleteDatasetOperator(
        task_id="delete_tabular_dataset",
        dataset_id=tabular_dataset_id,
        region=REGION,
        project_id=PROJECT_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_bucket = GCSDeleteBucketOperator(
        task_id="delete_bucket",
        bucket_name=DATA_SAMPLE_GCS_BUCKET_NAME,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    (
        # TEST SETUP
        create_bucket
        >> move_data_files
        >> download_training_script_file
        >> create_tabular_dataset
        >> create_custom_training_job
        >> create_custom_training_job_v2
        # TEST BODY
        >> get_model
        >> list_model_versions
        >> set_default_version
        >> add_version_alias
        >> upload_model
        >> upload_model_with_parent_model
        >> export_model
        >> delete_model
        >> delete_model_with_parent_model
        >> list_models
        # TEST TEARDOWN
        >> delete_version_alias
        >> delete_model_version
        >> delete_custom_training_job
        >> delete_tabular_dataset
        >> delete_bucket
    )

    # ### Everything below this line is not part of example ###
    # ### Just for system tests purpose ###
    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import asyncio
import json
import time
from collections.abc import MutableSequence, Sequence
from typing import TYPE_CHECKING, Any
from urllib.parse import urlencode, urljoin

from aiohttp import ClientSession
from google.api_core.client_options import ClientOptions
from google.api_core.gapic_v1.method import DEFAULT, _MethodDefault
from google.auth.transport.requests import AuthorizedSession, Request
from google.cloud.orchestration.airflow.service_v1 import (
    EnvironmentsAsyncClient,
    EnvironmentsClient,
    ImageVersionsClient,
    PollAirflowCommandResponse,
)

from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.google.common.consts import CLIENT_INFO
from airflow.providers.google.common.hooks.base_google import GoogleBaseAsyncHook, GoogleBaseHook

if TYPE_CHECKING:
    from google.api_core.operation import Operation
    from google.api_core.operation_async import AsyncOperation
    from google.api_core.retry import Retry
    from google.api_core.retry_async import AsyncRetry
    from google.cloud.orchestration.airflow.service_v1.services.environments.pagers import (
        ListEnvironmentsPager,
    )
    from google.cloud.orchestration.airflow.service_v1.services.image_versions.pagers import (
        ListImageVersionsPager,
    )
    from google.cloud.orchestration.airflow.service_v1.types import (
        Environment,
        ExecuteAirflowCommandResponse,
    )
    from google.protobuf.field_mask_pb2 import FieldMask

from airflow.providers.google.common.hooks.operation_helpers import OperationHelper


class CloudComposerHook(GoogleBaseHook, OperationHelper):
    """Hook for Google Cloud Composer APIs."""

    client_options = ClientOptions(api_endpoint="composer.googleapis.com:443")

    def get_environment_client(self) -> EnvironmentsClient:
        """Retrieve client library object that allow access Environments service."""
        return EnvironmentsClient(
            credentials=self.get_credentials(),
            client_info=CLIENT_INFO,
            client_options=self.client_options,
        )

    def get_image_versions_client(self) -> ImageVersionsClient:
        """Retrieve client library object that allow access Image Versions service."""
        return ImageVersionsClient(
            credentials=self.get_credentials(),
            client_info=CLIENT_INFO,
            client_options=self.client_options,
        )

    def make_composer_airflow_api_request(
        self,
        method: str,
        airflow_uri: str,
        path: str,
        data: Any | None = None,
        timeout: float | None = None,
    ):
        """
        Make a request to Cloud Composer environment's web server.

        :param method: The request method to use ('GET', 'OPTIONS', 'HEAD', 'POST', 'PUT', 'PATCH', 'DELETE').
        :param airflow_uri: The URI of the Apache Airflow Web UI hosted within this environment.
        :param path: The path to send the request.
        :param data: Dictionary, list of tuples, bytes, or file-like object to send in the body of the request.
        :param timeout: The timeout for this request.
        """
        authed_session = AuthorizedSession(self.get_credentials())

        resp = authed_session.request(
            method=method,
            url=urljoin(airflow_uri, path),
            data=data,
            headers={"Content-Type": "application/json"},
            timeout=timeout,
        )
        return resp

    def get_operation(self, operation_name):
        return self.get_environment_client().transport.operations_client.get_operation(name=operation_name)

    def get_environment_name(self, project_id, region, environment_id):
        return f"projects/{project_id}/locations/{region}/environments/{environment_id}"

    def get_parent(self, project_id, region):
        return f"projects/{project_id}/locations/{region}"

    @GoogleBaseHook.fallback_to_default_project_id
    def create_environment(
        self,
        project_id: str,
        region: str,
        environment: Environment | dict,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Operation:
        """
        Create a new environment.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param region: Required. The ID of the Google Cloud region that the service belongs to.
        :param environment:  The environment to create. This corresponds to the ``environment`` field on the
            ``request`` instance; if ``request`` is provided, this should not be set.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_environment_client()
        result = client.create_environment(
            request={"parent": self.get_parent(project_id, region), "environment": environment},
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def delete_environment(
        self,
        project_id: str,
        region: str,
        environment_id: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Operation:
        """
        Delete an environment.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param region: Required. The ID of the Google Cloud region that the service belongs to.
        :param environment_id: Required. The ID of the Google Cloud environment that the service belongs to.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_environment_client()
        name = self.get_environment_name(project_id, region, environment_id)
        result = client.delete_environment(
            request={"name": name}, retry=retry, timeout=timeout, metadata=metadata
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def get_environment(
        self,
        project_id: str,
        region: str,
        environment_id: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Environment:
        """
        Get an existing environment.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param region: Required. The ID of the Google Cloud region that the service belongs to.
        :param environment_id: Required. The ID of the Google Cloud environment that the service belongs to.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_environment_client()
        result = client.get_environment(
            request={"name": self.get_environment_name(project_id, region, environment_id)},
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def list_environments(
        self,
        project_id: str,
        region: str,
        page_size: int | None = None,
        page_token: str | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> ListEnvironmentsPager:
        """
        List environments.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param region: Required. The ID of the Google Cloud region that the service belongs to.
        :param page_size: The maximum number of environments to return.
        :param page_token: The next_page_token value returned from a previous List
            request, if any.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_environment_client()
        result = client.list_environments(
            request={
                "parent": self.get_parent(project_id, region),
                "page_size": page_size,
                "page_token": page_token,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def update_environment(
        self,
        project_id: str,
        region: str,
        environment_id: str,
        environment: Environment | dict,
        update_mask: dict | FieldMask,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Operation:
        r"""
        Update an environment.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param region: Required. The ID of the Google Cloud region that the service belongs to.
        :param environment_id: Required. The ID of the Google Cloud environment that the service belongs to.
        :param environment:  A patch environment. Fields specified by the ``updateMask`` will be copied from
            the patch environment into the environment under update.

            This corresponds to the ``environment`` field on the ``request`` instance; if ``request`` is
            provided, this should not be set.
        :param update_mask:  Required. A comma-separated list of paths, relative to ``Environment``, of fields
            to update. If a dict is provided, it must be of the same form as the protobuf message
            :class:`~google.protobuf.field_mask_pb2.FieldMask`
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_environment_client()
        name = self.get_environment_name(project_id, region, environment_id)

        result = client.update_environment(
            request={"name": name, "environment": environment, "update_mask": update_mask},
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def list_image_versions(
        self,
        project_id: str,
        region: str,
        page_size: int | None = None,
        page_token: str | None = None,
        include_past_releases: bool = False,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> ListImageVersionsPager:
        """
        List ImageVersions for provided location.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param region: Required. The ID of the Google Cloud region that the service belongs to.
        :param page_size: The maximum number of environments to return.
        :param page_token: The next_page_token value returned from a previous List
            request, if any.
        :param include_past_releases: Flag to include past releases
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_image_versions_client()
        result = client.list_image_versions(
            request={
                "parent": self.get_parent(project_id, region),
                "page_size": page_size,
                "page_token": page_token,
                "include_past_releases": include_past_releases,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def execute_airflow_command(
        self,
        project_id: str,
        region: str,
        environment_id: str,
        command: str,
        subcommand: str,
        parameters: MutableSequence[str],
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> ExecuteAirflowCommandResponse:
        """
        Execute Airflow command for provided Composer environment.

        :param project_id: The ID of the Google Cloud project that the service belongs to.
        :param region: The ID of the Google Cloud region that the service belongs to.
        :param environment_id: The ID of the Google Cloud environment that the service belongs to.
        :param command: Airflow command.
        :param subcommand: Airflow subcommand.
        :param parameters: Parameters for the Airflow command/subcommand as an array of arguments. It may
            contain positional arguments like ``["my-dag-id"]``, key-value parameters like ``["--foo=bar"]``
            or ``["--foo","bar"]``, or other flags like ``["-f"]``.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_environment_client()
        result = client.execute_airflow_command(
            request={
                "environment": self.get_environment_name(project_id, region, environment_id),
                "command": command,
                "subcommand": subcommand,
                "parameters": parameters,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    @GoogleBaseHook.fallback_to_default_project_id
    def poll_airflow_command(
        self,
        project_id: str,
        region: str,
        environment_id: str,
        execution_id: str,
        pod: str,
        pod_namespace: str,
        next_line_number: int,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> PollAirflowCommandResponse:
        """
        Poll Airflow command execution result for provided Composer environment.

        :param project_id: The ID of the Google Cloud project that the service belongs to.
        :param region: The ID of the Google Cloud region that the service belongs to.
        :param environment_id: The ID of the Google Cloud environment that the service belongs to.
        :param execution_id: The unique ID of the command execution.
        :param pod: The name of the pod where the command is executed.
        :param pod_namespace: The namespace of the pod where the command is executed.
        :param next_line_number: Line number from which new logs should be fetched.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = self.get_environment_client()
        result = client.poll_airflow_command(
            request={
                "environment": self.get_environment_name(project_id, region, environment_id),
                "execution_id": execution_id,
                "pod": pod,
                "pod_namespace": pod_namespace,
                "next_line_number": next_line_number,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )
        return result

    def wait_command_execution_result(
        self,
        project_id: str,
        region: str,
        environment_id: str,
        execution_cmd_info: dict,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        poll_interval: int = 10,
    ) -> dict:
        while True:
            try:
                result = self.poll_airflow_command(
                    project_id=project_id,
                    region=region,
                    environment_id=environment_id,
                    execution_id=execution_cmd_info["execution_id"],
                    pod=execution_cmd_info["pod"],
                    pod_namespace=execution_cmd_info["pod_namespace"],
                    next_line_number=1,
                    retry=retry,
                    timeout=timeout,
                    metadata=metadata,
                )
            except Exception as ex:
                self.log.exception("Exception occurred while polling CMD result")
                raise AirflowException(ex)

            result_dict = PollAirflowCommandResponse.to_dict(result)
            if result_dict["output_end"]:
                return result_dict

            self.log.info("Waiting for result...")
            time.sleep(poll_interval)

    def trigger_dag_run(
        self,
        composer_airflow_uri: str,
        composer_dag_id: str,
        composer_dag_conf: dict | None = None,
        timeout: float | None = None,
    ) -> dict:
        """
        Trigger DAG run for provided Apache Airflow Web UI hosted within Composer environment.

        :param composer_airflow_uri: The URI of the Apache Airflow Web UI hosted within Composer environment.
        :param composer_dag_id: The ID of DAG which will be triggered.
        :param composer_dag_conf: Configuration parameters for the DAG run.
        :param timeout: The timeout for this request.
        """
        response = self.make_composer_airflow_api_request(
            method="POST",
            airflow_uri=composer_airflow_uri,
            path=f"/api/v1/dags/{composer_dag_id}/dagRuns",
            data=json.dumps(
                {
                    "conf": composer_dag_conf or {},
                }
            ),
            timeout=timeout,
        )

        if response.status_code != 200:
            self.log.error(response.text)
            response.raise_for_status()

        return response.json()

    def get_dag_runs(
        self,
        composer_airflow_uri: str,
        composer_dag_id: str,
        timeout: float | None = None,
    ) -> dict:
        """
        Get the list of dag runs for provided DAG.

        :param composer_airflow_uri: The URI of the Apache Airflow Web UI hosted within Composer environment.
        :param composer_dag_id: The ID of DAG.
        :param timeout: The timeout for this request.
        """
        response = self.make_composer_airflow_api_request(
            method="GET",
            airflow_uri=composer_airflow_uri,
            path=f"/api/v1/dags/{composer_dag_id}/dagRuns",
            timeout=timeout,
        )

        if response.status_code != 200:
            self.log.error(
                "Failed to get DAG runs for dag_id=%s from %s (status=%s): %s",
                composer_dag_id,
                composer_airflow_uri,
                response.status_code,
                response.text,
            )
            response.raise_for_status()

        return response.json()

    def get_task_instances(
        self,
        composer_airflow_uri: str,
        composer_dag_id: str,
        query_parameters: dict | None = None,
        timeout: float | None = None,
    ) -> dict:
        """
        Get the list of task instances for provided DAG.

        :param composer_airflow_uri: The URI of the Apache Airflow Web UI hosted within Composer environment.
        :param composer_dag_id: The ID of DAG.
        :query_parameters: Query parameters for this request.
        :param timeout: The timeout for this request.
        """
        query_string = f"?{urlencode(query_parameters)}" if query_parameters else ""

        response = self.make_composer_airflow_api_request(
            method="GET",
            airflow_uri=composer_airflow_uri,
            path=f"/api/v1/dags/{composer_dag_id}/dagRuns/~/taskInstances{query_string}",
            timeout=timeout,
        )

        if response.status_code != 200:
            self.log.error(
                "Failed to get task instances for dag_id=%s from %s (status=%s): %s",
                composer_dag_id,
                composer_airflow_uri,
                response.status_code,
                response.text,
            )
            response.raise_for_status()

        return response.json()


class CloudComposerAsyncHook(GoogleBaseAsyncHook):
    """Hook for Google Cloud Composer async APIs."""

    sync_hook_class = CloudComposerHook

    client_options = ClientOptions(api_endpoint="composer.googleapis.com:443")

    async def get_environment_client(self) -> EnvironmentsAsyncClient:
        """Retrieve client library object that allow access Environments service."""
        sync_hook = await self.get_sync_hook()
        return EnvironmentsAsyncClient(
            credentials=sync_hook.get_credentials(),
            client_info=CLIENT_INFO,
            client_options=self.client_options,
        )

    async def make_composer_airflow_api_request(
        self,
        method: str,
        airflow_uri: str,
        path: str,
        data: Any | None = None,
        timeout: float | None = None,
    ):
        """
        Make a request to Cloud Composer environment's web server.

        :param method: The request method to use ('GET', 'OPTIONS', 'HEAD', 'POST', 'PUT', 'PATCH', 'DELETE').
        :param airflow_uri: The URI of the Apache Airflow Web UI hosted within this environment.
        :param path: The path to send the request.
        :param data: Dictionary, list of tuples, bytes, or file-like object to send in the body of the request.
        :param timeout: The timeout for this request.
        """
        sync_hook = await self.get_sync_hook()
        credentials = sync_hook.get_credentials()

        if not credentials.valid:
            credentials.refresh(Request())

        async with ClientSession() as session:
            async with session.request(
                method=method,
                url=urljoin(airflow_uri, path),
                data=data,
                headers={
                    "Content-Type": "application/json",
                    "Authorization": f"Bearer {credentials.token}",
                },
                timeout=timeout,
            ) as response:
                return await response.json(), response.status

    def get_environment_name(self, project_id, region, environment_id):
        return f"projects/{project_id}/locations/{region}/environments/{environment_id}"

    def get_parent(self, project_id, region):
        return f"projects/{project_id}/locations/{region}"

    async def get_operation(self, operation_name):
        client = await self.get_environment_client()
        return await client.transport.operations_client.get_operation(name=operation_name)

    @GoogleBaseHook.fallback_to_default_project_id
    async def create_environment(
        self,
        project_id: str,
        region: str,
        environment: Environment | dict,
        retry: AsyncRetry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> AsyncOperation:
        """
        Create a new environment.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param region: Required. The ID of the Google Cloud region that the service belongs to.
        :param environment:  The environment to create. This corresponds to the ``environment`` field on the
            ``request`` instance; if ``request`` is provided, this should not be set.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = await self.get_environment_client()
        return await client.create_environment(
            request={"parent": self.get_parent(project_id, region), "environment": environment},
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    @GoogleBaseHook.fallback_to_default_project_id
    async def delete_environment(
        self,
        project_id: str,
        region: str,
        environment_id: str,
        retry: AsyncRetry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> AsyncOperation:
        """
        Delete an environment.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param region: Required. The ID of the Google Cloud region that the service belongs to.
        :param environment_id: Required. The ID of the Google Cloud environment that the service belongs to.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = await self.get_environment_client()
        name = self.get_environment_name(project_id, region, environment_id)
        return await client.delete_environment(
            request={"name": name}, retry=retry, timeout=timeout, metadata=metadata
        )

    @GoogleBaseHook.fallback_to_default_project_id
    async def update_environment(
        self,
        project_id: str,
        region: str,
        environment_id: str,
        environment: Environment | dict,
        update_mask: dict | FieldMask,
        retry: AsyncRetry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> AsyncOperation:
        r"""
        Update an environment.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param region: Required. The ID of the Google Cloud region that the service belongs to.
        :param environment_id: Required. The ID of the Google Cloud environment that the service belongs to.
        :param environment:  A patch environment. Fields specified by the ``updateMask`` will be copied from
            the patch environment into the environment under update.

            This corresponds to the ``environment`` field on the ``request`` instance; if ``request`` is
            provided, this should not be set.
        :param update_mask:  Required. A comma-separated list of paths, relative to ``Environment``, of fields
            to update. If a dict is provided, it must be of the same form as the protobuf message
            :class:`~google.protobuf.field_mask_pb2.FieldMask`
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = await self.get_environment_client()
        name = self.get_environment_name(project_id, region, environment_id)

        return await client.update_environment(
            request={"name": name, "environment": environment, "update_mask": update_mask},
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    @GoogleBaseHook.fallback_to_default_project_id
    async def get_environment(
        self,
        project_id: str,
        region: str,
        environment_id: str,
        retry: AsyncRetry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Environment:
        """
        Get an existing environment.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param region: Required. The ID of the Google Cloud region that the service belongs to.
        :param environment_id: Required. The ID of the Google Cloud environment that the service belongs to.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = await self.get_environment_client()

        return await client.get_environment(
            request={"name": self.get_environment_name(project_id, region, environment_id)},
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    @GoogleBaseHook.fallback_to_default_project_id
    async def execute_airflow_command(
        self,
        project_id: str,
        region: str,
        environment_id: str,
        command: str,
        subcommand: str,
        parameters: MutableSequence[str],
        retry: AsyncRetry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> AsyncOperation:
        """
        Execute Airflow command for provided Composer environment.

        :param project_id: The ID of the Google Cloud project that the service belongs to.
        :param region: The ID of the Google Cloud region that the service belongs to.
        :param environment_id: The ID of the Google Cloud environment that the service belongs to.
        :param command: Airflow command.
        :param subcommand: Airflow subcommand.
        :param parameters: Parameters for the Airflow command/subcommand as an array of arguments. It may
            contain positional arguments like ``["my-dag-id"]``, key-value parameters like ``["--foo=bar"]``
            or ``["--foo","bar"]``, or other flags like ``["-f"]``.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = await self.get_environment_client()

        return await client.execute_airflow_command(
            request={
                "environment": self.get_environment_name(project_id, region, environment_id),
                "command": command,
                "subcommand": subcommand,
                "parameters": parameters,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    @GoogleBaseHook.fallback_to_default_project_id
    async def poll_airflow_command(
        self,
        project_id: str,
        region: str,
        environment_id: str,
        execution_id: str,
        pod: str,
        pod_namespace: str,
        next_line_number: int,
        retry: AsyncRetry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> AsyncOperation:
        """
        Poll Airflow command execution result for provided Composer environment.

        :param project_id: The ID of the Google Cloud project that the service belongs to.
        :param region: The ID of the Google Cloud region that the service belongs to.
        :param environment_id: The ID of the Google Cloud environment that the service belongs to.
        :param execution_id: The unique ID of the command execution.
        :param pod: The name of the pod where the command is executed.
        :param pod_namespace: The namespace of the pod where the command is executed.
        :param next_line_number: Line number from which new logs should be fetched.
        :param retry: Designation of what errors, if any, should be retried.
        :param timeout: The timeout for this request.
        :param metadata: Strings which should be sent along with the request as metadata.
        """
        client = await self.get_environment_client()

        return await client.poll_airflow_command(
            request={
                "environment": self.get_environment_name(project_id, region, environment_id),
                "execution_id": execution_id,
                "pod": pod,
                "pod_namespace": pod_namespace,
                "next_line_number": next_line_number,
            },
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    async def wait_command_execution_result(
        self,
        project_id: str,
        region: str,
        environment_id: str,
        execution_cmd_info: dict,
        retry: AsyncRetry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
        poll_interval: int = 10,
    ) -> dict:
        while True:
            try:
                result = await self.poll_airflow_command(
                    project_id=project_id,
                    region=region,
                    environment_id=environment_id,
                    execution_id=execution_cmd_info["execution_id"],
                    pod=execution_cmd_info["pod"],
                    pod_namespace=execution_cmd_info["pod_namespace"],
                    next_line_number=1,
                    retry=retry,
                    timeout=timeout,
                    metadata=metadata,
                )
            except Exception as ex:
                self.log.exception("Exception occurred while polling CMD result")
                raise AirflowException(ex)

            try:
                result_dict = PollAirflowCommandResponse.to_dict(result)
            except Exception as ex:
                self.log.exception("Exception occurred while transforming PollAirflowCommandResponse")
                raise AirflowException(ex)

            if result_dict["output_end"]:
                return result_dict

            self.log.info("Sleeping for %s seconds.", poll_interval)
            await asyncio.sleep(poll_interval)

    async def get_dag_runs(
        self,
        composer_airflow_uri: str,
        composer_dag_id: str,
        timeout: float | None = None,
    ) -> dict:
        """
        Get the list of dag runs for provided DAG.

        :param composer_airflow_uri: The URI of the Apache Airflow Web UI hosted within Composer environment.
        :param composer_dag_id: The ID of DAG.
        :param timeout: The timeout for this request.
        """
        response_body, response_status_code = await self.make_composer_airflow_api_request(
            method="GET",
            airflow_uri=composer_airflow_uri,
            path=f"/api/v1/dags/{composer_dag_id}/dagRuns",
            timeout=timeout,
        )

        if response_status_code != 200:
            self.log.error(
                "Failed to get DAG runs for dag_id=%s from %s (status=%s): %s",
                composer_dag_id,
                composer_airflow_uri,
                response_status_code,
                response_body["title"],
            )
            raise AirflowException(response_body["title"])

        return response_body

    async def get_task_instances(
        self,
        composer_airflow_uri: str,
        composer_dag_id: str,
        query_parameters: dict | None = None,
        timeout: float | None = None,
    ) -> dict:
        """
        Get the list of task instances for provided DAG.

        :param composer_airflow_uri: The URI of the Apache Airflow Web UI hosted within Composer environment.
        :param composer_dag_id: The ID of DAG.
        :query_parameters: Query parameters for this request.
        :param timeout: The timeout for this request.
        """
        query_string = f"?{urlencode(query_parameters)}" if query_parameters else ""

        response_body, response_status_code = await self.make_composer_airflow_api_request(
            method="GET",
            airflow_uri=composer_airflow_uri,
            path=f"/api/v1/dags/{composer_dag_id}/dagRuns/~/taskInstances{query_string}",
            timeout=timeout,
        )

        if response_status_code != 200:
            self.log.error(
                "Failed to get task instances for dag_id=%s from %s (status=%s): %s",
                composer_dag_id,
                composer_airflow_uri,
                response_status_code,
                response_body["title"],
            )
            raise AirflowException(response_body["title"])

        return response_body



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.providers.common.compat.sdk import AirflowOptionalProviderFeatureException

try:
    from airflow.providers.google.leveldb.hooks.leveldb import LevelDBHook
    from airflow.providers.google.leveldb.operators.leveldb import LevelDBOperator
except AirflowOptionalProviderFeatureException:
    pytest.skip("LevelDB not available", allow_module_level=True)


class TestLevelDBOperator:
    @mock.patch.dict("os.environ", AIRFLOW_CONN_LEVELDB_DEFAULT="test")
    @mock.patch.object(LevelDBHook, "run")
    def test_execute(self, mock_run):
        operator = LevelDBOperator(
            task_id="test_task",
            leveldb_conn_id="leveldb_default",
            command="put",
            key=b"key",
            value=b"value",
        )
        operator.execute(context="TEST_CONTEXT_ID")
        mock_run.assert_called_once_with(command="put", value=b"value", key=b"key", values=None, keys=None)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Utilities initializing and managing Go modules."""

from __future__ import annotations

import logging
import os
import shlex
import subprocess


def _execute_in_subprocess(cmd: list[str], cwd: str | None = None, env: dict[str, str] | None = None) -> None:
    """
    Execute a process and stream output to logger.

    :param cmd: command and arguments to run
    :param cwd: Current working directory passed to the Popen constructor
    :param env: Additional environment variables to set for the subprocess.
    """
    log = logging.getLogger(__name__)

    log.info("Executing cmd: %s", " ".join(shlex.quote(c) for c in cmd))
    with subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=0,
        close_fds=True,
        cwd=cwd,
        env=env,
    ) as proc:
        log.info("Output:")
        if proc.stdout:
            with proc.stdout:
                for line in iter(proc.stdout.readline, b""):
                    log.info("%s", line.decode().rstrip())

        exit_code = proc.wait()
    if exit_code != 0:
        raise subprocess.CalledProcessError(exit_code, cmd)


def init_module(go_module_name: str, go_module_path: str) -> None:
    """
    Initialize a Go module.

    If a ``go.mod`` file already exists, this function will do nothing.

    :param go_module_name: The name of the Go module to initialize.
    :param go_module_path: The path to the directory containing the Go module.
    """
    if os.path.isfile(os.path.join(go_module_path, "go.mod")):
        return
    go_mod_init_cmd = ["go", "mod", "init", go_module_name]
    _execute_in_subprocess(go_mod_init_cmd, cwd=go_module_path)


def install_dependencies(go_module_path: str) -> None:
    """
    Install dependencies for a Go module.

    :param go_module_path: The path to the directory containing the Go module.
    """
    go_mod_tidy = ["go", "mod", "tidy"]
    _execute_in_subprocess(go_mod_tidy, cwd=go_module_path)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

# For no Pydantic environment, we need to skip the tests
pytest.importorskip("google.cloud.aiplatform_v1")

from google.api_core.gapic_v1.method import DEFAULT
from google.cloud import aiplatform

# For no Pydantic environment, we need to skip the tests
pytest.importorskip("google.cloud.aiplatform_v1")

from google.cloud.aiplatform_v1 import JobState

from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.google.cloud.hooks.vertex_ai.hyperparameter_tuning_job import (
    HyperparameterTuningJobAsyncHook,
    HyperparameterTuningJobHook,
)

from unit.google.cloud.utils.base_gcp_mock import (
    mock_base_gcp_hook_default_project_id,
    mock_base_gcp_hook_no_default_project_id,
)

TEST_GCP_CONN_ID: str = "test-gcp-conn-id"
TEST_REGION: str = "test-region"
TEST_PROJECT_ID: str = "test-project-id"
TEST_HYPERPARAMETER_TUNING_JOB_ID = "test_hyperparameter_tuning_job_id"
TEST_UPDATE_MASK: dict = {}
TEST_DISPLAY_NAME = "Test display name"
TEST_METRIC_SPECS = {
    "accuracy": "maximize",
}
TEST_PARAM_SPECS = {
    "learning_rate": aiplatform.hyperparameter_tuning.DoubleParameterSpec(min=0.01, max=1, scale="log"),
}
TEST_MAX_TRIAL_COUNT = 3
TEST_PARALLEL_TRIAL_COUNT = 3
# CustomJob param
TEST_WORKER_POOL_SPECS = [
    {
        "machine_spec": {
            "machine_type": "n1-standard-4",
            "accelerator_type": "ACCELERATOR_TYPE_UNSPECIFIED",
            "accelerator_count": 0,
        },
    }
]
TEST_BASE_OUTPUT_DIR = None
TEST_CUSTOM_JOB_LABELS = None
TEST_CUSTOM_JOB_ENCRYPTION_SPEC_KEY_NAME = None
TEST_STAGING_BUCKET = None
# CustomJob param
TEST_MAX_FAILED_TRIAL_COUNT = 0
TEST_SEARCH_ALGORITHM = None
TEST_MEASUREMENT_SELECTION = "best"
TEST_HYPERPARAMETER_TUNING_JOB_LABELS = None
TEST_HYPERPARAMETER_TUNING_JOB_ENCRYPTION_SPEC_KEY_NAME = None
# run param
TEST_SERVICE_ACCOUNT = None
TEST_NETWORK = None
TEST_TIMEOUT = None
TEST_RESTART_JOB_ON_WORKER_RESTART = False
TEST_ENABLE_WEB_ACCESS = False
TEST_TENSORBOARD = None
TEST_SYNC = True

TEST_WAIT_JOB_COMPLETED = True
TEST_CREATE_HYPERPARAMETER_TUNING_JOB_PARAMS = dict(
    project_id=TEST_PROJECT_ID,
    region=TEST_REGION,
    display_name=TEST_DISPLAY_NAME,
    metric_spec=TEST_METRIC_SPECS,
    parameter_spec=TEST_PARAM_SPECS,
    max_trial_count=TEST_MAX_TRIAL_COUNT,
    parallel_trial_count=TEST_PARALLEL_TRIAL_COUNT,
    # START: CustomJob param
    worker_pool_specs=TEST_WORKER_POOL_SPECS,
    base_output_dir=TEST_BASE_OUTPUT_DIR,
    custom_job_labels=TEST_CUSTOM_JOB_LABELS,
    custom_job_encryption_spec_key_name=TEST_CUSTOM_JOB_ENCRYPTION_SPEC_KEY_NAME,
    staging_bucket=TEST_STAGING_BUCKET,
    # END: CustomJob param
    max_failed_trial_count=TEST_MAX_FAILED_TRIAL_COUNT,
    search_algorithm=TEST_SEARCH_ALGORITHM,
    measurement_selection=TEST_MEASUREMENT_SELECTION,
    hyperparameter_tuning_job_labels=TEST_HYPERPARAMETER_TUNING_JOB_LABELS,
    hyperparameter_tuning_job_encryption_spec_key_name=TEST_HYPERPARAMETER_TUNING_JOB_ENCRYPTION_SPEC_KEY_NAME,
    # START: run param
    service_account=TEST_SERVICE_ACCOUNT,
    network=TEST_NETWORK,
    timeout=TEST_TIMEOUT,
    restart_job_on_worker_restart=TEST_RESTART_JOB_ON_WORKER_RESTART,
    enable_web_access=TEST_ENABLE_WEB_ACCESS,
    tensorboard=TEST_TENSORBOARD,
    sync=TEST_SYNC,
    # END: run param
    wait_job_completed=TEST_WAIT_JOB_COMPLETED,
)
BASE_STRING = "airflow.providers.google.common.hooks.base_google.{}"
HYPERPARAMETER_TUNING_JOB_HOOK_PATH = (
    "airflow.providers.google.cloud.hooks.vertex_ai.hyperparameter_tuning_job.{}"
)
HYPERPARAMETER_TUNING_JOB_HOOK_STRING = (
    HYPERPARAMETER_TUNING_JOB_HOOK_PATH.format("HyperparameterTuningJobHook") + ".{}"
)
HYPERPARAMETER_TUNING_JOB_ASYNC_HOOK_STRING = (
    HYPERPARAMETER_TUNING_JOB_HOOK_PATH.format("HyperparameterTuningJobAsyncHook") + ".{}"
)


class TestHyperparameterTuningJobWithDefaultProjectIdHook:
    def setup_method(self):
        with mock.patch(
            BASE_STRING.format("GoogleBaseHook.__init__"), new=mock_base_gcp_hook_default_project_id
        ):
            self.hook = HyperparameterTuningJobHook(gcp_conn_id=TEST_GCP_CONN_ID)

    @mock.patch(HYPERPARAMETER_TUNING_JOB_HOOK_STRING.format("get_job_service_client"))
    def test_delete_hyperparameter_tuning_job(self, mock_client) -> None:
        self.hook.delete_hyperparameter_tuning_job(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            hyperparameter_tuning_job=TEST_HYPERPARAMETER_TUNING_JOB_ID,
        )
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.delete_hyperparameter_tuning_job.assert_called_once_with(
            request=dict(
                name=mock_client.return_value.hyperparameter_tuning_job_path.return_value,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )
        mock_client.return_value.hyperparameter_tuning_job_path.assert_called_once_with(
            TEST_PROJECT_ID,
            TEST_REGION,
            TEST_HYPERPARAMETER_TUNING_JOB_ID,
        )

    @mock.patch(HYPERPARAMETER_TUNING_JOB_HOOK_STRING.format("get_job_service_client"))
    def test_get_hyperparameter_tuning_job(self, mock_client) -> None:
        self.hook.get_hyperparameter_tuning_job(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            hyperparameter_tuning_job=TEST_HYPERPARAMETER_TUNING_JOB_ID,
        )
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.get_hyperparameter_tuning_job.assert_called_once_with(
            request=dict(
                name=mock_client.return_value.hyperparameter_tuning_job_path.return_value,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )
        mock_client.return_value.hyperparameter_tuning_job_path.assert_called_once_with(
            TEST_PROJECT_ID,
            TEST_REGION,
            TEST_HYPERPARAMETER_TUNING_JOB_ID,
        )

    @mock.patch(HYPERPARAMETER_TUNING_JOB_HOOK_STRING.format("get_job_service_client"))
    def test_list_hyperparameter_tuning_jobs(self, mock_client) -> None:
        self.hook.list_hyperparameter_tuning_jobs(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
        )
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.list_hyperparameter_tuning_jobs.assert_called_once_with(
            request=dict(
                parent=mock_client.return_value.common_location_path.return_value,
                filter=None,
                page_size=None,
                page_token=None,
                read_mask=None,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )
        mock_client.return_value.common_location_path.assert_called_once_with(TEST_PROJECT_ID, TEST_REGION)


class TestHyperparameterTuningJobWithoutDefaultProjectIdHook:
    def setup_method(self):
        with mock.patch(
            BASE_STRING.format("GoogleBaseHook.__init__"), new=mock_base_gcp_hook_no_default_project_id
        ):
            self.hook = HyperparameterTuningJobHook(gcp_conn_id=TEST_GCP_CONN_ID)

    @mock.patch(HYPERPARAMETER_TUNING_JOB_HOOK_STRING.format("get_job_service_client"))
    def test_delete_hyperparameter_tuning_job(self, mock_client) -> None:
        self.hook.delete_hyperparameter_tuning_job(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            hyperparameter_tuning_job=TEST_HYPERPARAMETER_TUNING_JOB_ID,
        )
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.delete_hyperparameter_tuning_job.assert_called_once_with(
            request=dict(
                name=mock_client.return_value.hyperparameter_tuning_job_path.return_value,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )
        mock_client.return_value.hyperparameter_tuning_job_path.assert_called_once_with(
            TEST_PROJECT_ID,
            TEST_REGION,
            TEST_HYPERPARAMETER_TUNING_JOB_ID,
        )

    @mock.patch(HYPERPARAMETER_TUNING_JOB_HOOK_STRING.format("get_job_service_client"))
    def test_get_hyperparameter_tuning_job(self, mock_client) -> None:
        self.hook.get_hyperparameter_tuning_job(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
            hyperparameter_tuning_job=TEST_HYPERPARAMETER_TUNING_JOB_ID,
        )
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.get_hyperparameter_tuning_job.assert_called_once_with(
            request=dict(
                name=mock_client.return_value.hyperparameter_tuning_job_path.return_value,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )
        mock_client.return_value.hyperparameter_tuning_job_path.assert_called_once_with(
            TEST_PROJECT_ID,
            TEST_REGION,
            TEST_HYPERPARAMETER_TUNING_JOB_ID,
        )

    @mock.patch(HYPERPARAMETER_TUNING_JOB_HOOK_STRING.format("get_job_service_client"))
    def test_list_hyperparameter_tuning_jobs(self, mock_client) -> None:
        self.hook.list_hyperparameter_tuning_jobs(
            project_id=TEST_PROJECT_ID,
            region=TEST_REGION,
        )
        mock_client.assert_called_once_with(TEST_REGION)
        mock_client.return_value.list_hyperparameter_tuning_jobs.assert_called_once_with(
            request=dict(
                parent=mock_client.return_value.common_location_path.return_value,
                filter=None,
                page_size=None,
                page_token=None,
                read_mask=None,
            ),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )
        mock_client.return_value.common_location_path.assert_called_once_with(TEST_PROJECT_ID, TEST_REGION)


class TestHyperparameterTuningJobHook:
    def setup_method(self):
        with mock.patch(
            BASE_STRING.format("GoogleBaseHook.__init__"), new=mock_base_gcp_hook_default_project_id
        ):
            self.hook = HyperparameterTuningJobHook(gcp_conn_id=TEST_GCP_CONN_ID)

    @mock.patch(HYPERPARAMETER_TUNING_JOB_HOOK_STRING.format("get_hyperparameter_tuning_job_object"))
    @mock.patch(HYPERPARAMETER_TUNING_JOB_HOOK_STRING.format("get_custom_job_object"))
    def test_create_hyperparameter_tuning_job(
        self,
        mock_get_custom_job_object,
        mock_get_hyperparameter_tuning_job_object,
    ):
        mock_custom_job = mock_get_custom_job_object.return_value

        result = self.hook.create_hyperparameter_tuning_job(**TEST_CREATE_HYPERPARAMETER_TUNING_JOB_PARAMS)

        mock_get_custom_job_object.assert_called_once_with(
            project=TEST_PROJECT_ID,
            location=TEST_REGION,
            display_name=TEST_DISPLAY_NAME,
            worker_pool_specs=TEST_WORKER_POOL_SPECS,
            base_output_dir=TEST_BASE_OUTPUT_DIR,
            labels=TEST_CUSTOM_JOB_LABELS,
            encryption_spec_key_name=TEST_CUSTOM_JOB_ENCRYPTION_SPEC_KEY_NAME,
            staging_bucket=TEST_STAGING_BUCKET,
        )
        mock_get_hyperparameter_tuning_job_object.assert_called_once_with(
            project=TEST_PROJECT_ID,
            location=TEST_REGION,
            display_name=TEST_DISPLAY_NAME,
            custom_job=mock_custom_job,
            metric_spec=TEST_METRIC_SPECS,
            parameter_spec=TEST_PARAM_SPECS,
            max_trial_count=TEST_MAX_TRIAL_COUNT,
            parallel_trial_count=TEST_PARALLEL_TRIAL_COUNT,
            max_failed_trial_count=TEST_MAX_FAILED_TRIAL_COUNT,
            search_algorithm=TEST_SEARCH_ALGORITHM,
            measurement_selection=TEST_MEASUREMENT_SELECTION,
            labels=TEST_HYPERPARAMETER_TUNING_JOB_LABELS,
            encryption_spec_key_name=TEST_HYPERPARAMETER_TUNING_JOB_ENCRYPTION_SPEC_KEY_NAME,
        )
        self.hook._hyperparameter_tuning_job.run.assert_called_once_with(
            service_account=TEST_SERVICE_ACCOUNT,
            network=TEST_NETWORK,
            timeout=TEST_TIMEOUT,
            restart_job_on_worker_restart=TEST_RESTART_JOB_ON_WORKER_RESTART,
            enable_web_access=TEST_ENABLE_WEB_ACCESS,
            tensorboard=TEST_TENSORBOARD,
            sync=TEST_SYNC,
        )
        self.hook._hyperparameter_tuning_job.wait.assert_called_once()
        self.hook._hyperparameter_tuning_job._wait_for_resource_creation.assert_not_called()
        assert result == self.hook._hyperparameter_tuning_job

    @mock.patch(HYPERPARAMETER_TUNING_JOB_HOOK_STRING.format("get_hyperparameter_tuning_job_object"))
    @mock.patch(HYPERPARAMETER_TUNING_JOB_HOOK_STRING.format("get_custom_job_object"))
    def test_create_hyperparameter_tuning_job_no_wait(self, _, __):
        params = dict(TEST_CREATE_HYPERPARAMETER_TUNING_JOB_PARAMS)
        params["wait_job_completed"] = False

        result = self.hook.create_hyperparameter_tuning_job(**params)

        self.hook._hyperparameter_tuning_job.wait.assert_not_called()
        self.hook._hyperparameter_tuning_job._wait_for_resource_creation.assert_called_once()
        assert result == self.hook._hyperparameter_tuning_job


class TestHyperparameterTuningJobAsyncHook:
    def setup_method(self):
        with mock.patch(
            BASE_STRING.format("GoogleBaseHook.__init__"), new=mock_base_gcp_hook_default_project_id
        ):
            self.hook = HyperparameterTuningJobAsyncHook(gcp_conn_id=TEST_GCP_CONN_ID)

    @pytest.mark.asyncio
    @mock.patch(HYPERPARAMETER_TUNING_JOB_ASYNC_HOOK_STRING.format("get_job_service_client"))
    async def test_get_hyperparameter_tuning_job(self, mock_get_job_service_client):
        mock_client = mock.MagicMock()
        mock_get_job_service_client.side_effect = mock.AsyncMock(return_value=mock_client)
        mock_job_name = mock_client.hyperparameter_tuning_job_path.return_value
        mock_job = mock.MagicMock()
        mock_async_get_hyperparameter_tuning_job = mock.AsyncMock(return_value=mock_job)
        mock_client.get_hyperparameter_tuning_job.side_effect = mock_async_get_hyperparameter_tuning_job

        result = await self.hook.get_hyperparameter_tuning_job(
            project_id=TEST_PROJECT_ID,
            location=TEST_REGION,
            job_id=TEST_HYPERPARAMETER_TUNING_JOB_ID,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

        mock_get_job_service_client.assert_called_once_with(region=TEST_REGION)
        mock_client.hyperparameter_tuning_job_path.assert_called_once_with(
            TEST_PROJECT_ID, TEST_REGION, TEST_HYPERPARAMETER_TUNING_JOB_ID
        )
        mock_async_get_hyperparameter_tuning_job.assert_awaited_once_with(
            request={"name": mock_job_name},
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )
        assert result == mock_job

    @pytest.mark.parametrize(
        "state",
        [
            JobState.JOB_STATE_CANCELLED,
            JobState.JOB_STATE_FAILED,
            JobState.JOB_STATE_PAUSED,
            JobState.JOB_STATE_SUCCEEDED,
        ],
    )
    @pytest.mark.asyncio
    @mock.patch(HYPERPARAMETER_TUNING_JOB_HOOK_PATH.format("asyncio.sleep"))
    async def test_wait_hyperparameter_tuning_job(self, mock_sleep, state):
        mock_job = mock.MagicMock(state=state)
        mock_async_get_hpt_job = mock.AsyncMock(return_value=mock_job)
        mock_get_hpt_job = mock.MagicMock(side_effect=mock_async_get_hpt_job)

        await_kwargs = dict(
            project_id=TEST_PROJECT_ID,
            location=TEST_REGION,
            job_id=TEST_HYPERPARAMETER_TUNING_JOB_ID,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )
        with mock.patch.object(self.hook, "get_hyperparameter_tuning_job", mock_get_hpt_job):
            result = await self.hook.wait_hyperparameter_tuning_job(**await_kwargs)

        mock_async_get_hpt_job.assert_awaited_once_with(**await_kwargs)
        mock_sleep.assert_not_awaited()
        assert result == mock_job

    @pytest.mark.parametrize(
        "state",
        [
            JobState.JOB_STATE_UNSPECIFIED,
            JobState.JOB_STATE_QUEUED,
            JobState.JOB_STATE_PENDING,
            JobState.JOB_STATE_RUNNING,
            JobState.JOB_STATE_CANCELLING,
            JobState.JOB_STATE_EXPIRED,
            JobState.JOB_STATE_UPDATING,
            JobState.JOB_STATE_PARTIALLY_SUCCEEDED,
        ],
    )
    @pytest.mark.asyncio
    @mock.patch(HYPERPARAMETER_TUNING_JOB_HOOK_PATH.format("asyncio.sleep"))
    async def test_wait_hyperparameter_tuning_job_waited(self, mock_sleep, state):
        mock_job_incomplete = mock.MagicMock(state=state)
        mock_job_complete = mock.MagicMock(state=JobState.JOB_STATE_SUCCEEDED)
        mock_async_get_ht_job = mock.AsyncMock(side_effect=[mock_job_incomplete, mock_job_complete])
        mock_get_ht_job = mock.MagicMock(side_effect=mock_async_get_ht_job)

        await_kwargs = dict(
            project_id=TEST_PROJECT_ID,
            location=TEST_REGION,
            job_id=TEST_HYPERPARAMETER_TUNING_JOB_ID,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )

        with mock.patch.object(self.hook, "get_hyperparameter_tuning_job", mock_get_ht_job):
            result = await self.hook.wait_hyperparameter_tuning_job(**await_kwargs)

        mock_async_get_ht_job.assert_has_awaits(
            [
                mock.call(**await_kwargs),
                mock.call(**await_kwargs),
            ]
        )
        mock_sleep.assert_awaited_once()
        assert result == mock_job_complete

    @pytest.mark.asyncio
    async def test_wait_hyperparameter_tuning_job_exception(self):
        mock_get_ht_job = mock.MagicMock(side_effect=Exception)
        with mock.patch.object(self.hook, "get_hyperparameter_tuning_job", mock_get_ht_job):
            with pytest.raises(AirflowException):
                await self.hook.wait_hyperparameter_tuning_job(
                    project_id=TEST_PROJECT_ID,
                    location=TEST_REGION,
                    job_id=TEST_HYPERPARAMETER_TUNING_JOB_ID,
                    retry=DEFAULT,
                    timeout=None,
                    metadata=(),
                )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Hooks for Google Analytics (GA4) Admin service.

.. spelling:word-list::

    DataStream
    ListAccountsPager
    ListGoogleAdsLinksPager
"""

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

from google.analytics.admin_v1beta import (
    AnalyticsAdminServiceClient,
    DataStream,
    Property,
)
from google.api_core.gapic_v1.method import DEFAULT, _MethodDefault

from airflow.providers.google.common.consts import CLIENT_INFO
from airflow.providers.google.common.hooks.base_google import GoogleBaseHook

if TYPE_CHECKING:
    from google.analytics.admin_v1beta.services.analytics_admin_service.pagers import (
        ListAccountsPager,
        ListGoogleAdsLinksPager,
    )
    from google.api_core.retry import Retry


class GoogleAnalyticsAdminHook(GoogleBaseHook):
    """Hook for Google Analytics 4 (GA4) Admin API."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._conn: AnalyticsAdminServiceClient | None = None

    def get_conn(self) -> AnalyticsAdminServiceClient:
        if not self._conn:
            self._conn = AnalyticsAdminServiceClient(
                credentials=self.get_credentials(), client_info=CLIENT_INFO
            )
        return self._conn

    def list_accounts(
        self,
        page_size: int | None = None,
        page_token: str | None = None,
        show_deleted: bool | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> ListAccountsPager:
        """
        Get list of accounts in Google Analytics.

        .. seealso::
            For more details please check the client library documentation:
            https://developers.google.com/analytics/devguides/config/admin/v1/rest/v1beta/accounts/list

        :param page_size: Optional, number of results to return in the list.
        :param page_token: Optional. The next_page_token value returned from a previous List request, if any.
        :param show_deleted: Optional. Whether to include soft-deleted (ie: "trashed") Accounts in the results.
        :param retry: Optional, a retry object used  to retry requests. If `None` is specified, requests
            will not be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.

        :returns: List of Google Analytics accounts.
        """
        request = {"page_size": page_size, "page_token": page_token, "show_deleted": show_deleted}
        client = self.get_conn()
        return client.list_accounts(request=request, retry=retry, timeout=timeout, metadata=metadata)

    def create_property(
        self,
        analytics_property: Property | dict,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Property:
        """
        Create Google Analytics property.

        .. seealso::
            For more details please check the client library documentation:
            https://developers.google.com/analytics/devguides/config/admin/v1/rest/v1beta/properties/create

        :param analytics_property: The property to create. Note: the supplied property must specify its
            parent.
        :param retry: Optional, a retry object used  to retry requests. If `None` is specified, requests
            will not be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.

        :returns: Created Google Analytics property.
        """
        client = self.get_conn()
        return client.create_property(
            request={"property": analytics_property},
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    def delete_property(
        self,
        property_id: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> Property:
        """
        Soft delete Google Analytics property.

        .. seealso::
            For more details please check the client library documentation:
            https://developers.google.com/analytics/devguides/config/admin/v1/rest/v1beta/properties/delete

        :param property_id: ID of the Property to soft-delete. Format: properties/{property_id}.
        :param retry: Optional, a retry object used  to retry requests. If `None` is specified, requests
            will not be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.

        :returns: Resource message representing Google Analytics property.
        """
        client = self.get_conn()
        request = {"name": f"properties/{property_id}"}
        return client.delete_property(request=request, retry=retry, timeout=timeout, metadata=metadata)

    def create_data_stream(
        self,
        property_id: str,
        data_stream: DataStream | dict,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> DataStream:
        """
        Create Google Analytics data stream.

        .. seealso::
            For more details please check the client library documentation:
            https://developers.google.com/analytics/devguides/config/admin/v1/rest/v1beta/properties.dataStreams/create

        :param property_id: ID of the parent property for the data stream.
        :param data_stream: The data stream to create.
        :param retry: Optional, a retry object used  to retry requests. If `None` is specified, requests
            will not be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.

        :returns: Created Google Analytics data stream.
        """
        client = self.get_conn()
        return client.create_data_stream(
            request={"parent": f"properties/{property_id}", "data_stream": data_stream},
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    def delete_data_stream(
        self,
        property_id: str,
        data_stream_id: str,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> None:
        """
        Delete Google Analytics data stream.

        .. seealso::
            For more details please check the client library documentation:
            https://developers.google.com/analytics/devguides/config/admin/v1/rest/v1beta/properties.dataStreams/delete

        :param property_id: ID of the parent property for the data stream.
        :param data_stream_id: The data stream id to delete.
        :param retry: Optional, a retry object used  to retry requests. If `None` is specified, requests
            will not be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.
        """
        client = self.get_conn()
        return client.delete_data_stream(
            request={"name": f"properties/{property_id}/dataStreams/{data_stream_id}"},
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    def list_google_ads_links(
        self,
        property_id: str,
        page_size: int | None = None,
        page_token: str | None = None,
        retry: Retry | _MethodDefault = DEFAULT,
        timeout: float | None = None,
        metadata: Sequence[tuple[str, str]] = (),
    ) -> ListGoogleAdsLinksPager:
        """
        Get list of Google Ads links.

        .. seealso::
            For more details please check the client library documentation:
            https://googleapis.dev/python/analyticsadmin/latest/admin_v1beta/analytics_admin_service.html#google.analytics.admin_v1beta.services.analytics_admin_service.AnalyticsAdminServiceAsyncClient.list_google_ads_links

        :param property_id: ID of the parent property.
        :param page_size: Optional, number of results to return in the list.
        :param page_token: Optional. The next_page_token value returned from a previous List request, if any.
        :param retry: Optional, a retry object used  to retry requests. If `None` is specified, requests
            will not be retried.
        :param timeout: Optional. The timeout for this request.
        :param metadata: Optional. Strings which should be sent along with the request as metadata.

        :returns: List of Google Analytics accounts.
        """
        client = self.get_conn()
        request = {"parent": f"properties/{property_id}", "page_size": page_size, "page_token": page_token}
        return client.list_google_ads_links(request=request, retry=retry, timeout=timeout, metadata=metadata)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import logging
import re
from unittest import mock

import pytest
from google.api_core.exceptions import NotFound
from google.cloud.secretmanager_v1.types import AccessSecretVersionResponse

from airflow.models import Connection
from airflow.providers.common.compat.sdk import AirflowException
from airflow.providers.google.cloud.secrets.secret_manager import CloudSecretManagerBackend

CREDENTIALS = "test-creds"
KEY_FILE = "test-file.json"
PROJECT_ID = "test-project-id"
OVERRIDDEN_PROJECT_ID = "overridden-test-project-id"
CONNECTIONS_PREFIX = "test-connections"
VARIABLES_PREFIX = "test-variables"
SEP = "-"
CONN_ID = "test-postgres"
CONN_URI = "postgresql://airflow:airflow@host:5432/airflow"
VAR_KEY = "hello"
VAR_VALUE = "world"
CONFIG_KEY = "sql_alchemy_conn"
CONFIG_VALUE = "postgresql://airflow:airflow@host:5432/airflow"

MODULE_NAME = "airflow.providers.google.cloud.secrets.secret_manager"
CLIENT_MODULE_NAME = "airflow.providers.google.cloud._internal_client.secret_manager_client"


class TestCloudSecretManagerBackend:
    @mock.patch(MODULE_NAME + ".get_credentials_and_project_id")
    @mock.patch(CLIENT_MODULE_NAME + ".SecretManagerServiceClient")
    def test_default_valid_and_sep(self, mock_client_callable, mock_get_creds):
        mock_get_creds.return_value = CREDENTIALS, PROJECT_ID
        mock_client = mock.MagicMock()
        mock_client_callable.return_value = mock_client

        backend = CloudSecretManagerBackend()
        assert backend._is_valid_prefix_and_sep()

    @pytest.mark.parametrize(
        ("prefix", "sep"),
        [
            pytest.param("not:valid", ":", id="colon separator"),
            pytest.param("not/valid", "/", id="backslash separator"),
            pytest.param("a b", "", id="space with char and empty separator"),
            pytest.param(" ", "", id="space only and empty separator"),
        ],
    )
    def test_raise_exception_with_invalid_prefix_sep(self, prefix, sep):
        with pytest.raises(AirflowException):
            CloudSecretManagerBackend(connections_prefix=prefix, sep=sep)

    @pytest.mark.parametrize(
        ("prefix", "sep", "is_valid"),
        [
            pytest.param("valid1", "-", True, id="valid: dash separator"),
            pytest.param("isValid", "_", True, id="valid: underscore separator"),
            pytest.param("", "", True, id="valid: empty string and empty separator"),
            pytest.param("", " ", False, id="invalid: empty string and space separator"),
            pytest.param("not:valid", ":", False, id="invalid: colon separator"),
        ],
    )
    @mock.patch(MODULE_NAME + ".get_credentials_and_project_id")
    @mock.patch(CLIENT_MODULE_NAME + ".SecretManagerServiceClient")
    def test_is_valid_prefix_and_sep(self, mock_client_callable, mock_get_creds, prefix, sep, is_valid):
        mock_get_creds.return_value = CREDENTIALS, PROJECT_ID
        mock_client = mock.MagicMock()
        mock_client_callable.return_value = mock_client
        backend = CloudSecretManagerBackend()
        backend.connections_prefix = prefix
        backend.sep = sep
        assert backend._is_valid_prefix_and_sep() == is_valid

    @pytest.mark.parametrize("connections_prefix", ["airflow-connections", "connections", "airflow"])
    @mock.patch(MODULE_NAME + ".get_credentials_and_project_id")
    @mock.patch(CLIENT_MODULE_NAME + ".SecretManagerServiceClient")
    def test_get_conn_uri(self, mock_client_callable, mock_get_creds, connections_prefix):
        mock_get_creds.return_value = CREDENTIALS, PROJECT_ID
        mock_client = mock.MagicMock()
        mock_client_callable.return_value = mock_client

        test_response = AccessSecretVersionResponse()
        test_response.payload.data = CONN_URI.encode("UTF-8")
        mock_client.access_secret_version.return_value = test_response

        secrets_manager_backend = CloudSecretManagerBackend(connections_prefix=connections_prefix)
        secret_id = secrets_manager_backend.build_path(connections_prefix, CONN_ID, SEP)
        returned_uri = secrets_manager_backend.get_conn_value(conn_id=CONN_ID)
        assert returned_uri == CONN_URI
        mock_client.secret_version_path.assert_called_once_with(PROJECT_ID, secret_id, "latest")

    @mock.patch(MODULE_NAME + ".get_credentials_and_project_id")
    @mock.patch(MODULE_NAME + ".CloudSecretManagerBackend.get_conn_value")
    def test_get_connection(self, mock_get_value, mock_get_creds):
        mock_get_creds.return_value = CREDENTIALS, PROJECT_ID
        mock_get_value.return_value = CONN_URI
        conn = CloudSecretManagerBackend().get_connection(conn_id=CONN_ID)
        assert isinstance(conn, Connection)

    @mock.patch(MODULE_NAME + ".get_credentials_and_project_id")
    @mock.patch(CLIENT_MODULE_NAME + ".SecretManagerServiceClient")
    def test_get_conn_uri_non_existent_key(self, mock_client_callable, mock_get_creds, caplog):
        mock_get_creds.return_value = CREDENTIALS, PROJECT_ID
        mock_client = mock.MagicMock()
        mock_client_callable.return_value = mock_client
        # The requested secret id or secret version does not exist
        mock_client.access_secret_version.side_effect = NotFound("test-msg")

        secrets_manager_backend = CloudSecretManagerBackend(connections_prefix=CONNECTIONS_PREFIX)
        secret_id = secrets_manager_backend.build_path(CONNECTIONS_PREFIX, CONN_ID, SEP)
        with caplog.at_level(level=logging.DEBUG, logger=secrets_manager_backend.client.log.name):
            assert secrets_manager_backend.get_conn_value(conn_id=CONN_ID) is None
            assert secrets_manager_backend.get_connection(conn_id=CONN_ID) is None
            assert re.search(
                f"Google Cloud API Call Error \\(NotFound\\): Secret ID {secret_id} not found",
                caplog.messages[0],
            )

    @pytest.mark.parametrize("variables_prefix", ["airflow-variables", "variables", "airflow"])
    @mock.patch(MODULE_NAME + ".get_credentials_and_project_id")
    @mock.patch(CLIENT_MODULE_NAME + ".SecretManagerServiceClient")
    def test_get_variable(self, mock_client_callable, mock_get_creds, variables_prefix):
        mock_get_creds.return_value = CREDENTIALS, PROJECT_ID
        mock_client = mock.MagicMock()
        mock_client_callable.return_value = mock_client

        test_response = AccessSecretVersionResponse()
        test_response.payload.data = VAR_VALUE.encode("UTF-8")
        mock_client.access_secret_version.return_value = test_response

        secrets_manager_backend = CloudSecretManagerBackend(variables_prefix=variables_prefix)
        secret_id = secrets_manager_backend.build_path(variables_prefix, VAR_KEY, SEP)
        returned_uri = secrets_manager_backend.get_variable(VAR_KEY)
        assert returned_uri == VAR_VALUE
        mock_client.secret_version_path.assert_called_once_with(PROJECT_ID, secret_id, "latest")

    @pytest.mark.parametrize("config_prefix", ["airflow-config", "config", "airflow"])
    @mock.patch(MODULE_NAME + ".get_credentials_and_project_id")
    @mock.patch(CLIENT_MODULE_NAME + ".SecretManagerServiceClient")
    def test_get_config(self, mock_client_callable, mock_get_creds, config_prefix):
        mock_get_creds.return_value = CREDENTIALS, PROJECT_ID
        mock_client = mock.MagicMock()
        mock_client_callable.return_value = mock_client

        test_response = AccessSecretVersionResponse()
        test_response.payload.data = CONFIG_VALUE.encode("UTF-8")
        mock_client.access_secret_version.return_value = test_response

        secrets_manager_backend = CloudSecretManagerBackend(config_prefix=config_prefix)
        secret_id = secrets_manager_backend.build_path(config_prefix, CONFIG_KEY, SEP)
        returned_val = secrets_manager_backend.get_config(CONFIG_KEY)
        assert returned_val == CONFIG_VALUE
        mock_client.secret_version_path.assert_called_once_with(PROJECT_ID, secret_id, "latest")

    @pytest.mark.parametrize("variables_prefix", ["airflow-variables", "variables", "airflow"])
    @mock.patch(MODULE_NAME + ".get_credentials_and_project_id")
    @mock.patch(CLIENT_MODULE_NAME + ".SecretManagerServiceClient")
    def test_get_variable_override_project_id(self, mock_client_callable, mock_get_creds, variables_prefix):
        mock_get_creds.return_value = CREDENTIALS, PROJECT_ID
        mock_client = mock.MagicMock()
        mock_client_callable.return_value = mock_client

        test_response = AccessSecretVersionResponse()
        test_response.payload.data = VAR_VALUE.encode("UTF-8")
        mock_client.access_secret_version.return_value = test_response

        secrets_manager_backend = CloudSecretManagerBackend(
            variables_prefix=variables_prefix, project_id=OVERRIDDEN_PROJECT_ID
        )
        secret_id = secrets_manager_backend.build_path(variables_prefix, VAR_KEY, SEP)
        returned_uri = secrets_manager_backend.get_variable(VAR_KEY)
        assert returned_uri == VAR_VALUE
        mock_client.secret_version_path.assert_called_once_with(OVERRIDDEN_PROJECT_ID, secret_id, "latest")

    @mock.patch(MODULE_NAME + ".get_credentials_and_project_id")
    @mock.patch(CLIENT_MODULE_NAME + ".SecretManagerServiceClient")
    def test_get_variable_non_existent_key(self, mock_client_callable, mock_get_creds, caplog):
        mock_get_creds.return_value = CREDENTIALS, PROJECT_ID
        mock_client = mock.MagicMock()
        mock_client_callable.return_value = mock_client
        # The requested secret id or secret version does not exist
        mock_client.access_secret_version.side_effect = NotFound("test-msg")

        secrets_manager_backend = CloudSecretManagerBackend(variables_prefix=VARIABLES_PREFIX)
        secret_id = secrets_manager_backend.build_path(VARIABLES_PREFIX, VAR_KEY, SEP)
        with caplog.at_level(level=logging.DEBUG, logger=secrets_manager_backend.client.log.name):
            assert secrets_manager_backend.get_variable(VAR_KEY) is None
            assert re.search(
                f"Google Cloud API Call Error \\(NotFound\\): Secret ID {secret_id} not found",
                caplog.messages[0],
            )

    @mock.patch(MODULE_NAME + ".get_credentials_and_project_id")
    @mock.patch(CLIENT_MODULE_NAME + ".SecretManagerServiceClient")
    def test_connections_prefix_none_value(self, mock_client_callable, mock_get_creds):
        mock_get_creds.return_value = CREDENTIALS, PROJECT_ID
        mock_client = mock.MagicMock()
        mock_client_callable.return_value = mock_client

        with mock.patch(MODULE_NAME + ".CloudSecretManagerBackend._get_secret") as mock_get_secret:
            with mock.patch(
                MODULE_NAME + ".CloudSecretManagerBackend._is_valid_prefix_and_sep"
            ) as mock_is_valid_prefix_sep:
                secrets_manager_backend = CloudSecretManagerBackend(connections_prefix=None)

                mock_is_valid_prefix_sep.assert_not_called()
                assert secrets_manager_backend.get_conn_value(conn_id=CONN_ID) is None
                mock_get_secret.assert_not_called()

    @mock.patch(MODULE_NAME + ".get_credentials_and_project_id")
    @mock.patch(CLIENT_MODULE_NAME + ".SecretManagerServiceClient")
    def test_variables_prefix_none_value(self, mock_client_callable, mock_get_creds):
        mock_get_creds.return_value = CREDENTIALS, PROJECT_ID
        mock_client = mock.MagicMock()
        mock_client_callable.return_value = mock_client

        with mock.patch(MODULE_NAME + ".CloudSecretManagerBackend._get_secret") as mock_get_secret:
            secrets_manager_backend = CloudSecretManagerBackend(variables_prefix=None)

            assert secrets_manager_backend.get_variable(VAR_KEY) is None
            mock_get_secret.assert_not_called()

    @mock.patch(MODULE_NAME + ".get_credentials_and_project_id")
    @mock.patch(CLIENT_MODULE_NAME + ".SecretManagerServiceClient")
    def test_config_prefix_none_value(self, mock_client_callable, mock_get_creds):
        mock_get_creds.return_value = CREDENTIALS, PROJECT_ID
        mock_client = mock.MagicMock()
        mock_client_callable.return_value = mock_client

        with mock.patch(MODULE_NAME + ".CloudSecretManagerBackend._get_secret") as mock_get_secret:
            secrets_manager_backend = CloudSecretManagerBackend(config_prefix=None)

            assert secrets_manager_backend.get_config(CONFIG_KEY) is None
            mock_get_secret.assert_not_called()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


"""
Example Airflow DAG for Google Vertex AI service testing Auto ML operations.
"""

from __future__ import annotations

import os
from datetime import datetime

from google.cloud.aiplatform import schema
from google.protobuf.json_format import ParseDict
from google.protobuf.struct_pb2 import Value

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.vertex_ai.auto_ml import (
    CreateAutoMLForecastingTrainingJobOperator,
    DeleteAutoMLTrainingJobOperator,
)
from airflow.providers.google.cloud.operators.vertex_ai.dataset import (
    CreateDatasetOperator,
    DeleteDatasetOperator,
)

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT", "default")
DAG_ID = "vertex_ai_auto_ml_operations"
REGION = "us-central1"
FORECASTING_DISPLAY_NAME = f"auto-ml-forecasting-{ENV_ID}"
MODEL_DISPLAY_NAME = f"auto-ml-forecasting-model-{ENV_ID}"

RESOURCE_DATA_BUCKET = "airflow-system-tests-resources"

FORECAST_DATASET = {
    "display_name": f"forecast-dataset-{ENV_ID}",
    "metadata_schema_uri": schema.dataset.metadata.time_series,
    "metadata": ParseDict(
        {
            "input_config": {
                "gcs_source": {
                    "uri": [f"gs://{RESOURCE_DATA_BUCKET}/vertex-ai/datasets/forecast-dataset.csv"]
                }
            }
        },
        Value(),
    ),
}

TEST_TIME_COLUMN = "date"
TEST_TIME_SERIES_IDENTIFIER_COLUMN = "store_name"
TEST_TARGET_COLUMN = "sale_dollars"

COLUMN_SPECS = {
    TEST_TIME_COLUMN: "timestamp",
    TEST_TARGET_COLUMN: "numeric",
    "city": "categorical",
    "zip_code": "categorical",
    "county": "categorical",
}


with DAG(
    f"{DAG_ID}_forecasting_training_job",
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "vertex_ai", "auto_ml"],
) as dag:
    create_forecast_dataset = CreateDatasetOperator(
        task_id="forecast_dataset",
        dataset=FORECAST_DATASET,
        region=REGION,
        project_id=PROJECT_ID,
    )
    forecast_dataset_id = create_forecast_dataset.output["dataset_id"]

    # [START how_to_cloud_vertex_ai_create_auto_ml_forecasting_training_job_operator]
    create_auto_ml_forecasting_training_job = CreateAutoMLForecastingTrainingJobOperator(
        task_id="auto_ml_forecasting_task",
        display_name=FORECASTING_DISPLAY_NAME,
        optimization_objective="minimize-rmse",
        column_specs=COLUMN_SPECS,
        # run params
        dataset_id=forecast_dataset_id,
        target_column=TEST_TARGET_COLUMN,
        time_column=TEST_TIME_COLUMN,
        time_series_identifier_column=TEST_TIME_SERIES_IDENTIFIER_COLUMN,
        available_at_forecast_columns=[TEST_TIME_COLUMN],
        unavailable_at_forecast_columns=[TEST_TARGET_COLUMN],
        time_series_attribute_columns=["city", "zip_code", "county"],
        forecast_horizon=30,
        context_window=30,
        data_granularity_unit="day",
        data_granularity_count=1,
        weight_column=None,
        budget_milli_node_hours=1000,
        model_display_name=MODEL_DISPLAY_NAME,
        predefined_split_column_name=None,
        region=REGION,
        project_id=PROJECT_ID,
    )
    # [END how_to_cloud_vertex_ai_create_auto_ml_forecasting_training_job_operator]

    # [START how_to_cloud_vertex_ai_delete_auto_ml_training_job_operator]
    delete_auto_ml_forecasting_training_job = DeleteAutoMLTrainingJobOperator(
        task_id="delete_auto_ml_forecasting_training_job",
        training_pipeline_id="{{ task_instance.xcom_pull(task_ids='auto_ml_forecasting_task', "
        "key='training_id') }}",
        region=REGION,
        project_id=PROJECT_ID,
    )
    # [END how_to_cloud_vertex_ai_delete_auto_ml_training_job_operator]

    delete_forecast_dataset = DeleteDatasetOperator(
        task_id="delete_forecast_dataset",
        dataset_id=forecast_dataset_id,
        region=REGION,
        project_id=PROJECT_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    (
        # TEST SETUP
        create_forecast_dataset
        # TEST BODY
        >> create_auto_ml_forecasting_training_job
        # TEST TEARDOWN
        >> delete_auto_ml_forecasting_training_job
        >> delete_forecast_dataset
    )

    # ### Everything below this line is not part of example ###
    # ### Just for system tests purpose ###
    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from airflow.providers.google.cloud.transfers.azure_blob_to_gcs import AzureBlobStorageToGCSOperator

WASB_CONN_ID = "wasb_default"
GCP_CONN_ID = "google_cloud_default"
BLOB_NAME = "azure_blob"
CONTAINER_NAME = "azure_container"
BUCKET_NAME = "airflow"
OBJECT_NAME = "file.txt"
FILENAME = "file.txt"
GZIP = False
IMPERSONATION_CHAIN = None
TASK_ID = "transfer_file"


class TestAzureBlobStorageToGCSTransferOperator:
    def test_init(self):
        operator = AzureBlobStorageToGCSOperator(
            wasb_conn_id=WASB_CONN_ID,
            gcp_conn_id=GCP_CONN_ID,
            blob_name=BLOB_NAME,
            container_name=CONTAINER_NAME,
            bucket_name=BUCKET_NAME,
            object_name=OBJECT_NAME,
            filename=FILENAME,
            gzip=GZIP,
            impersonation_chain=IMPERSONATION_CHAIN,
            task_id=TASK_ID,
        )
        assert operator.wasb_conn_id == WASB_CONN_ID
        assert operator.blob_name == BLOB_NAME
        assert operator.container_name == CONTAINER_NAME
        assert operator.gcp_conn_id == GCP_CONN_ID
        assert operator.bucket_name == BUCKET_NAME
        assert operator.object_name == OBJECT_NAME
        assert operator.filename == FILENAME
        assert operator.gzip == GZIP
        assert operator.impersonation_chain == IMPERSONATION_CHAIN
        assert operator.task_id == TASK_ID

    @mock.patch("airflow.providers.google.cloud.transfers.azure_blob_to_gcs.WasbHook")
    @mock.patch("airflow.providers.google.cloud.transfers.azure_blob_to_gcs.GCSHook")
    @mock.patch("airflow.providers.google.cloud.transfers.azure_blob_to_gcs.tempfile")
    def test_execute(self, mock_temp, mock_hook_gcs, mock_hook_wasb):
        op = AzureBlobStorageToGCSOperator(
            wasb_conn_id=WASB_CONN_ID,
            gcp_conn_id=GCP_CONN_ID,
            blob_name=BLOB_NAME,
            container_name=CONTAINER_NAME,
            bucket_name=BUCKET_NAME,
            object_name=OBJECT_NAME,
            filename=FILENAME,
            gzip=GZIP,
            impersonation_chain=IMPERSONATION_CHAIN,
            task_id=TASK_ID,
        )

        op.execute(context=None)
        mock_hook_wasb.assert_called_once_with(wasb_conn_id=WASB_CONN_ID)

        mock_hook_wasb.return_value.get_file.assert_called_once_with(
            file_path=mock_temp.NamedTemporaryFile.return_value.__enter__.return_value.name,
            container_name=CONTAINER_NAME,
            blob_name=BLOB_NAME,
        )
        mock_hook_gcs.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID, impersonation_chain=IMPERSONATION_CHAIN
        )
        mock_hook_gcs.return_value.upload.assert_called_once_with(
            bucket_name=BUCKET_NAME,
            object_name=OBJECT_NAME,
            gzip=GZIP,
            filename=mock_temp.NamedTemporaryFile.return_value.__enter__.return_value.name,
        )

    @mock.patch("airflow.providers.google.cloud.transfers.azure_blob_to_gcs.WasbHook")
    def test_execute_single_file_transfer_openlineage(self, mock_hook_wasb):
        from airflow.providers.common.compat.openlineage.facet import Dataset

        MOCK_AZURE_ACCOUNT_NAME = "mock_account_name"
        mock_hook_wasb.return_value.get_conn.return_value.account_name = MOCK_AZURE_ACCOUNT_NAME

        operator = AzureBlobStorageToGCSOperator(
            wasb_conn_id=WASB_CONN_ID,
            gcp_conn_id=GCP_CONN_ID,
            blob_name=BLOB_NAME,
            container_name=CONTAINER_NAME,
            bucket_name=BUCKET_NAME,
            object_name=OBJECT_NAME,
            filename=FILENAME,
            gzip=GZIP,
            impersonation_chain=IMPERSONATION_CHAIN,
            task_id=TASK_ID,
        )

        lineage = operator.get_openlineage_facets_on_start()

        assert len(lineage.inputs) == 1
        assert len(lineage.outputs) == 1
        assert lineage.inputs[0] == Dataset(
            namespace=f"wasbs://{CONTAINER_NAME}@{MOCK_AZURE_ACCOUNT_NAME}", name=BLOB_NAME
        )
        assert lineage.outputs[0] == Dataset(namespace=f"gs://{BUCKET_NAME}", name=OBJECT_NAME)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import MagicMock

import pytest
from google.cloud.bigquery.retry import DEFAULT_RETRY
from google.cloud.bigquery.table import Table

from airflow.providers.common.compat.openlineage.facet import (
    ColumnLineageDatasetFacet,
    Dataset,
    DocumentationDatasetFacet,
    ExternalQueryRunFacet,
    Fields,
    Identifier,
    InputField,
    SchemaDatasetFacet,
    SchemaDatasetFacetFields,
    SymlinksDatasetFacet,
)
from airflow.providers.common.compat.sdk import TaskDeferred
from airflow.providers.google.cloud.transfers.bigquery_to_gcs import BigQueryToGCSOperator
from airflow.providers.google.cloud.triggers.bigquery import BigQueryInsertJobTrigger

TASK_ID = "test-bq-create-table-operator"
TEST_DATASET = "test-dataset"
TEST_TABLE_ID = "test-table-id"
PROJECT_ID = "test-project-id"
JOB_PROJECT_ID = "job-project-id"
TEST_BUCKET = "test-bucket"
TEST_FOLDER = "test-folder"
TEST_OBJECT_NO_WILDCARD = "file.extension"
TEST_OBJECT_WILDCARD = "file_*.extension"
TEST_TABLE_API_REPR = {
    "tableReference": {"projectId": PROJECT_ID, "datasetId": TEST_DATASET, "tableId": TEST_TABLE_ID},
    "description": "Table description.",
    "schema": {
        "fields": [
            {"name": "field1", "type": "STRING", "description": "field1 description"},
            {"name": "field2", "type": "INTEGER"},
        ]
    },
}
TEST_TABLE: Table = Table.from_api_repr(TEST_TABLE_API_REPR)
TEST_EMPTY_TABLE_API_REPR = {
    "tableReference": {"projectId": PROJECT_ID, "datasetId": TEST_DATASET, "tableId": TEST_TABLE_ID}
}
TEST_EMPTY_TABLE: Table = Table.from_api_repr(TEST_EMPTY_TABLE_API_REPR)


class TestBigQueryToGCSOperator:
    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_gcs.BigQueryHook")
    def test_execute(self, mock_hook):
        source_project_dataset_table = f"{PROJECT_ID}:{TEST_DATASET}.{TEST_TABLE_ID}"
        destination_cloud_storage_uris = ["gs://some-bucket/some-file.txt"]
        compression = "NONE"
        export_format = "CSV"
        field_delimiter = ","
        print_header = True
        labels = {"k1": "v1"}
        job_id = "123456"
        hash_ = "hash"
        real_job_id = f"{job_id}_{hash_}"

        expected_configuration = {
            "extract": {
                "sourceTable": {
                    "projectId": "test-project-id",
                    "datasetId": "test-dataset",
                    "tableId": "test-table-id",
                },
                "compression": "NONE",
                "destinationUris": ["gs://some-bucket/some-file.txt"],
                "destinationFormat": "CSV",
                "fieldDelimiter": ",",
                "printHeader": True,
            },
            "labels": {"k1": "v1"},
        }

        mock_hook.return_value.split_tablename.return_value = (PROJECT_ID, TEST_DATASET, TEST_TABLE_ID)
        mock_hook.return_value.generate_job_id.return_value = real_job_id
        mock_hook.return_value.insert_job.return_value = MagicMock(job_id="real_job_id", error_result=False)
        mock_hook.return_value.project_id = JOB_PROJECT_ID

        operator = BigQueryToGCSOperator(
            task_id=TASK_ID,
            source_project_dataset_table=source_project_dataset_table,
            destination_cloud_storage_uris=destination_cloud_storage_uris,
            compression=compression,
            export_format=export_format,
            field_delimiter=field_delimiter,
            print_header=print_header,
            labels=labels,
            project_id=JOB_PROJECT_ID,
        )
        result = operator.execute(context=mock.MagicMock())

        assert result is not None
        assert isinstance(result, list)
        assert result == ["gs://some-bucket/some-file.txt"]

        mock_hook.return_value.insert_job.assert_called_once_with(
            job_id="123456_hash",
            configuration=expected_configuration,
            project_id=JOB_PROJECT_ID,
            location=None,
            timeout=None,
            retry=DEFAULT_RETRY,
            nowait=False,
        )

    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_gcs.BigQueryHook")
    def test_execute_deferrable_mode(self, mock_hook):
        source_project_dataset_table = f"{PROJECT_ID}:{TEST_DATASET}.{TEST_TABLE_ID}"
        destination_cloud_storage_uris = ["gs://some-bucket/some-file.txt"]
        compression = "NONE"
        export_format = "CSV"
        field_delimiter = ","
        print_header = True
        labels = {"k1": "v1"}
        job_id = "123456"
        hash_ = "hash"
        real_job_id = f"{job_id}_{hash_}"

        expected_configuration = {
            "extract": {
                "sourceTable": {
                    "projectId": "test-project-id",
                    "datasetId": "test-dataset",
                    "tableId": "test-table-id",
                },
                "compression": "NONE",
                "destinationUris": ["gs://some-bucket/some-file.txt"],
                "destinationFormat": "CSV",
                "fieldDelimiter": ",",
                "printHeader": True,
            },
            "labels": {"k1": "v1"},
        }

        mock_hook.return_value.split_tablename.return_value = (PROJECT_ID, TEST_DATASET, TEST_TABLE_ID)
        mock_hook.return_value.generate_job_id.return_value = real_job_id
        mock_hook.return_value.insert_job.return_value = MagicMock(job_id="real_job_id", error_result=False)
        mock_hook.return_value.project_id = JOB_PROJECT_ID

        operator = BigQueryToGCSOperator(
            project_id=JOB_PROJECT_ID,
            task_id=TASK_ID,
            source_project_dataset_table=source_project_dataset_table,
            destination_cloud_storage_uris=destination_cloud_storage_uris,
            compression=compression,
            export_format=export_format,
            field_delimiter=field_delimiter,
            print_header=print_header,
            labels=labels,
            deferrable=True,
        )
        with pytest.raises(TaskDeferred) as exc:
            operator.execute(context=mock.MagicMock())

        assert isinstance(exc.value.trigger, BigQueryInsertJobTrigger), (
            "Trigger is not a BigQueryInsertJobTrigger"
        )

        mock_hook.return_value.insert_job.assert_called_once_with(
            configuration=expected_configuration,
            job_id="123456_hash",
            project_id=JOB_PROJECT_ID,
            location=None,
            timeout=None,
            retry=DEFAULT_RETRY,
            nowait=True,
        )

    def test_execute_complete_reassigns_job_id(self):
        """Assert that we use job_id from event after deferral."""

        operator = BigQueryToGCSOperator(
            project_id=JOB_PROJECT_ID,
            task_id=TASK_ID,
            source_project_dataset_table=f"{PROJECT_ID}.{TEST_DATASET}.{TEST_TABLE_ID}",
            destination_cloud_storage_uris=[f"gs://{TEST_BUCKET}/{TEST_FOLDER}/"],
            deferrable=True,
            job_id=None,
        )
        job_id = "123456"

        assert operator.job_id is None
        operator.execute_complete(
            context=MagicMock(),
            event={"status": "success", "message": "Job completed", "job_id": job_id},
        )
        assert operator.job_id == job_id

    def test_execute_complete_returns_destination_cloud_storage_uris(self):
        """Assert that self.destination_cloud_storage_uris is returned by execute_complete."""

        operator = BigQueryToGCSOperator(
            project_id=JOB_PROJECT_ID,
            task_id=TASK_ID,
            source_project_dataset_table=f"{PROJECT_ID}.{TEST_DATASET}.{TEST_TABLE_ID}",
            destination_cloud_storage_uris=[f"gs://{TEST_BUCKET}/{TEST_FOLDER}/"],
            deferrable=True,
            job_id=None,
        )

        result = operator.execute_complete(
            context=MagicMock(),
            event={"status": "success", "message": "Job completed", "job_id": None},
        )

        assert result is not None
        assert isinstance(result, list)
        assert result == [f"gs://{TEST_BUCKET}/{TEST_FOLDER}/"]

    @pytest.mark.parametrize(
        ("gcs_uri", "expected_dataset_name"),
        (
            (
                f"gs://{TEST_BUCKET}/{TEST_FOLDER}/{TEST_OBJECT_NO_WILDCARD}",
                f"{TEST_FOLDER}/{TEST_OBJECT_NO_WILDCARD}",
            ),
            (f"gs://{TEST_BUCKET}/{TEST_OBJECT_NO_WILDCARD}", TEST_OBJECT_NO_WILDCARD),
            (f"gs://{TEST_BUCKET}/{TEST_FOLDER}/{TEST_OBJECT_WILDCARD}", TEST_FOLDER),
            (f"gs://{TEST_BUCKET}/{TEST_OBJECT_WILDCARD}", "/"),
            (f"gs://{TEST_BUCKET}/{TEST_FOLDER}/*", TEST_FOLDER),
        ),
    )
    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_gcs.BigQueryHook")
    def test_get_openlineage_facets_on_complete_gcs_dataset_name(
        self, mock_hook, gcs_uri, expected_dataset_name
    ):
        operator = BigQueryToGCSOperator(
            project_id=JOB_PROJECT_ID,
            task_id=TASK_ID,
            source_project_dataset_table=f"{PROJECT_ID}.{TEST_DATASET}.{TEST_TABLE_ID}",
            destination_cloud_storage_uris=[gcs_uri],
        )

        mock_hook.return_value.split_tablename.return_value = (PROJECT_ID, TEST_DATASET, TEST_TABLE_ID)
        operator.execute(context=mock.MagicMock())

        lineage = operator.get_openlineage_facets_on_complete(None)
        assert len(lineage.outputs) == 1
        assert lineage.outputs[0].name == expected_dataset_name

    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_gcs.BigQueryHook")
    def test_get_openlineage_facets_on_complete_gcs_multiple_uris(self, mock_hook):
        operator = BigQueryToGCSOperator(
            project_id=JOB_PROJECT_ID,
            task_id=TASK_ID,
            source_project_dataset_table=f"{PROJECT_ID}.{TEST_DATASET}.{TEST_TABLE_ID}",
            destination_cloud_storage_uris=[
                f"gs://{TEST_BUCKET}1/{TEST_FOLDER}1/{TEST_OBJECT_NO_WILDCARD}",
                f"gs://{TEST_BUCKET}2/{TEST_FOLDER}2/{TEST_OBJECT_WILDCARD}",
                f"gs://{TEST_BUCKET}3/{TEST_OBJECT_NO_WILDCARD}",
                f"gs://{TEST_BUCKET}4/{TEST_OBJECT_WILDCARD}",
            ],
        )

        mock_hook.return_value.split_tablename.return_value = (PROJECT_ID, TEST_DATASET, TEST_TABLE_ID)
        operator.execute(context=mock.MagicMock())

        lineage = operator.get_openlineage_facets_on_complete(None)
        assert len(lineage.outputs) == 4
        assert lineage.outputs[0].name == f"{TEST_FOLDER}1/{TEST_OBJECT_NO_WILDCARD}"
        assert lineage.outputs[1].name == f"{TEST_FOLDER}2"
        assert lineage.outputs[2].name == TEST_OBJECT_NO_WILDCARD
        assert lineage.outputs[3].name == "/"

    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_gcs.BigQueryHook")
    def test_get_openlineage_facets_on_complete_bq_dataset(self, mock_hook):
        source_project_dataset_table = f"{PROJECT_ID}.{TEST_DATASET}.{TEST_TABLE_ID}"

        expected_input_dataset_facets = {
            "schema": SchemaDatasetFacet(
                fields=[
                    SchemaDatasetFacetFields(name="field1", type="STRING", description="field1 description"),
                    SchemaDatasetFacetFields(name="field2", type="INTEGER"),
                ]
            ),
            "documentation": DocumentationDatasetFacet(description="Table description."),
        }

        mock_hook.return_value.split_tablename.return_value = (PROJECT_ID, TEST_DATASET, TEST_TABLE_ID)
        mock_hook.return_value.get_client.return_value.get_table.return_value = TEST_TABLE

        operator = BigQueryToGCSOperator(
            project_id=JOB_PROJECT_ID,
            task_id=TASK_ID,
            source_project_dataset_table=source_project_dataset_table,
            destination_cloud_storage_uris=["gs://bucket/file"],
        )
        operator.execute(context=mock.MagicMock())

        lineage = operator.get_openlineage_facets_on_complete(None)
        assert len(lineage.inputs) == 1
        assert lineage.inputs[0] == Dataset(
            namespace="bigquery",
            name=source_project_dataset_table,
            facets=expected_input_dataset_facets,
        )

    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_gcs.BigQueryHook")
    def test_get_openlineage_facets_on_complete_bq_dataset_empty_table(self, mock_hook):
        source_project_dataset_table = f"{PROJECT_ID}.{TEST_DATASET}.{TEST_TABLE_ID}"

        mock_hook.return_value.split_tablename.return_value = (PROJECT_ID, TEST_DATASET, TEST_TABLE_ID)
        mock_hook.return_value.get_client.return_value.get_table.return_value = TEST_EMPTY_TABLE

        operator = BigQueryToGCSOperator(
            project_id=JOB_PROJECT_ID,
            task_id=TASK_ID,
            source_project_dataset_table=source_project_dataset_table,
            destination_cloud_storage_uris=["gs://bucket/file"],
        )
        operator.execute(context=mock.MagicMock())

        lineage = operator.get_openlineage_facets_on_complete(None)
        assert len(lineage.inputs) == 1
        assert lineage.inputs[0] == Dataset(
            namespace="bigquery",
            name=source_project_dataset_table,
            facets={},
        )

    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_gcs.BigQueryHook")
    def test_get_openlineage_facets_on_complete_gcs_no_wildcard_empty_table(self, mock_hook):
        source_project_dataset_table = f"{PROJECT_ID}.{TEST_DATASET}.{TEST_TABLE_ID}"
        destination_cloud_storage_uris = [f"gs://{TEST_BUCKET}/{TEST_FOLDER}/{TEST_OBJECT_NO_WILDCARD}"]
        real_job_id = "123456_hash"
        bq_namespace = "bigquery"

        mock_hook.return_value.split_tablename.return_value = (PROJECT_ID, TEST_DATASET, TEST_TABLE_ID)
        mock_hook.return_value.insert_job.return_value = MagicMock(job_id=real_job_id, error_result=False)
        mock_hook.return_value.get_client.return_value.get_table.return_value = TEST_EMPTY_TABLE

        operator = BigQueryToGCSOperator(
            project_id=JOB_PROJECT_ID,
            task_id=TASK_ID,
            source_project_dataset_table=source_project_dataset_table,
            destination_cloud_storage_uris=destination_cloud_storage_uris,
        )

        operator.execute(context=mock.MagicMock())

        lineage = operator.get_openlineage_facets_on_complete(None)
        assert len(lineage.inputs) == 1
        assert len(lineage.outputs) == 1
        assert lineage.inputs[0] == Dataset(
            namespace=bq_namespace, name=source_project_dataset_table, facets={}
        )
        assert lineage.outputs[0] == Dataset(
            namespace=f"gs://{TEST_BUCKET}",
            name=f"{TEST_FOLDER}/{TEST_OBJECT_NO_WILDCARD}",
            facets={},
        )
        assert lineage.run_facets == {
            "externalQuery": ExternalQueryRunFacet(externalQueryId=real_job_id, source=bq_namespace)
        }
        assert lineage.job_facets == {}

    @mock.patch("airflow.providers.google.cloud.transfers.bigquery_to_gcs.BigQueryHook")
    def test_get_openlineage_facets_on_complete_gcs_wildcard_full_table(self, mock_hook):
        source_project_dataset_table = f"{PROJECT_ID}.{TEST_DATASET}.{TEST_TABLE_ID}"
        destination_cloud_storage_uris = [f"gs://{TEST_BUCKET}/{TEST_FOLDER}/{TEST_OBJECT_WILDCARD}"]
        real_job_id = "123456_hash"
        bq_namespace = "bigquery"

        schema_facet = SchemaDatasetFacet(
            fields=[
                SchemaDatasetFacetFields(name="field1", type="STRING", description="field1 description"),
                SchemaDatasetFacetFields(name="field2", type="INTEGER"),
            ]
        )
        expected_input_facets = {
            "schema": schema_facet,
            "documentation": DocumentationDatasetFacet(description="Table description."),
        }

        expected_output_facets = {
            "schema": schema_facet,
            "columnLineage": ColumnLineageDatasetFacet(
                fields={
                    "field1": Fields(
                        inputFields=[
                            InputField(
                                namespace=bq_namespace, name=source_project_dataset_table, field="field1"
                            )
                        ],
                        transformationType="IDENTITY",
                        transformationDescription="identical",
                    ),
                    "field2": Fields(
                        inputFields=[
                            InputField(
                                namespace=bq_namespace, name=source_project_dataset_table, field="field2"
                            )
                        ],
                        transformationType="IDENTITY",
                        transformationDescription="identical",
                    ),
                }
            ),
            "symlink": SymlinksDatasetFacet(
                identifiers=[
                    Identifier(
                        namespace=f"gs://{TEST_BUCKET}",
                        name=f"{TEST_FOLDER}/{TEST_OBJECT_WILDCARD}",
                        type="file",
                    )
                ]
            ),
        }

        mock_hook.return_value.split_tablename.return_value = (PROJECT_ID, TEST_DATASET, TEST_TABLE_ID)
        mock_hook.return_value.insert_job.return_value = MagicMock(job_id=real_job_id, error_result=False)
        mock_hook.return_value.get_client.return_value.get_table.return_value = TEST_TABLE

        operator = BigQueryToGCSOperator(
            project_id=JOB_PROJECT_ID,
            task_id=TASK_ID,
            source_project_dataset_table=source_project_dataset_table,
            destination_cloud_storage_uris=destination_cloud_storage_uris,
        )

        operator.execute(context=mock.MagicMock())

        lineage = operator.get_openlineage_facets_on_complete(None)
        assert len(lineage.inputs) == 1
        assert len(lineage.outputs) == 1
        assert lineage.inputs[0] == Dataset(
            namespace=bq_namespace, name=source_project_dataset_table, facets=expected_input_facets
        )
        assert lineage.outputs[0] == Dataset(
            namespace=f"gs://{TEST_BUCKET}", name=TEST_FOLDER, facets=expected_output_facets
        )
        assert lineage.run_facets == {
            "externalQuery": ExternalQueryRunFacet(externalQueryId=real_job_id, source=bq_namespace)
        }
        assert lineage.job_facets == {}



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

pytest.importorskip("wtforms")  # Remove after upgrading to FAB5


class TestValidJson:
    def setup_method(self):
        self.form_field_mock = mock.MagicMock(data='{"valid":"True"}')
        self.form_field_mock.gettext.side_effect = lambda msg: msg
        self.form_mock = mock.MagicMock(spec_set=dict)

    def _validate(self, message=None):
        from airflow.providers.google.cloud.utils.validators import ValidJson

        validator = ValidJson(message=message)

        return validator(self.form_mock, self.form_field_mock)

    def test_form_field_is_none(self):
        self.form_field_mock.data = None

        assert self._validate() is None

    def test_validation_pass(self):
        assert self._validate() is None

    def test_validation_raises_default_message(self):
        from wtforms.validators import ValidationError

        self.form_field_mock.data = "2017-05-04"

        with pytest.raises(ValidationError, match="JSON Validation Error:.*"):
            self._validate()

    def test_validation_raises_custom_message(self):
        from wtforms.validators import ValidationError

        self.form_field_mock.data = "2017-05-04"

        with pytest.raises(ValidationError, match="Invalid JSON"):
            self._validate(
                message="Invalid JSON: {}",
            )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import os
from datetime import datetime

from airflow.models.dag import DAG
from airflow.providers.google.cloud.transfers.azure_blob_to_gcs import AzureBlobStorageToGCSOperator
from airflow.providers.microsoft.azure.sensors.wasb import (
    WasbBlobSensor,
)

BLOB_NAME = os.environ.get("AZURE_BLOB_NAME", "file.txt")
AZURE_CONTAINER_NAME = os.environ.get("AZURE_CONTAINER_NAME", "airflow")
GCP_BUCKET_FILE_PATH = os.environ.get("GCP_BUCKET_FILE_PATH", "file.txt")
GCP_BUCKET_NAME = os.environ.get("GCP_BUCKET_NAME", "INVALID BUCKET NAME")
GCP_OBJECT_NAME = os.environ.get("GCP_OBJECT_NAME", "file.txt")
DAG_ID = "example_azure_blob_to_gcs"

with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),  # Override to match your needs
) as dag:
    wait_for_blob = WasbBlobSensor(
        task_id="wait_for_blob", container_name=AZURE_CONTAINER_NAME, blob_name=BLOB_NAME
    )

    # [START how_to_azure_blob_to_gcs]
    transfer_files_to_gcs = AzureBlobStorageToGCSOperator(
        task_id="transfer_files_to_gcs",
        # azure args
        container_name=AZURE_CONTAINER_NAME,
        blob_name=BLOB_NAME,
        # GCP args
        bucket_name=GCP_BUCKET_NAME,
        object_name=GCP_OBJECT_NAME,
        filename=GCP_BUCKET_FILE_PATH,
        gzip=False,
        impersonation_chain=None,
    )
    # [END how_to_azure_blob_to_gcs]

    (wait_for_blob >> transfer_files_to_gcs)

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)
